#  Copyright 2015 Odin
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import copy
import errno
import os
from unittest import mock

import ddt
from os_brick.remotefs import remotefs
from oslo_utils import units

from cinder import context
from cinder import exception
from cinder.image import image_utils
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit import test
from cinder.volume.drivers import vzstorage


_orig_path_exists = os.path.exists


@ddt.ddt
class VZStorageTestCase(test.TestCase):

    _FAKE_SHARE = "10.0.0.1,10.0.0.2:/cluster123:123123"
    _FAKE_MNT_BASE = '/mnt'
    _FAKE_MNT_POINT = os.path.join(_FAKE_MNT_BASE, 'fake_hash')
    _FAKE_VOLUME_NAME = 'volume-4f711859-4928-4cb7-801a-a50c37ceaccc'
    _FAKE_VOLUME_PATH = os.path.join(_FAKE_MNT_POINT, _FAKE_VOLUME_NAME)
    _FAKE_SNAPSHOT_ID = '50811859-4928-4cb7-801a-a50c37ceacba'
    _FAKE_SNAPSHOT_PATH = (
        _FAKE_VOLUME_PATH + '-snapshot' + _FAKE_SNAPSHOT_ID)

    def setUp(self):
        super(VZStorageTestCase, self).setUp()

        self._cfg = mock.MagicMock()
        self._cfg.vzstorage_shares_config = '/fake/config/path'
        self._cfg.vzstorage_sparsed_volumes = False
        self._cfg.vzstorage_used_ratio = 0.7
        self._cfg.vzstorage_mount_point_base = self._FAKE_MNT_BASE
        self._cfg.vzstorage_default_volume_format = 'raw'
        self._cfg.nas_secure_file_operations = 'auto'
        self._cfg.nas_secure_file_permissions = 'auto'

        self._vz_driver = vzstorage.VZStorageDriver(configuration=self._cfg)
        self._vz_driver._local_volume_dir = mock.Mock(
            return_value=self._FAKE_MNT_POINT)
        self._vz_driver._execute = mock.Mock()
        self._vz_driver.base = self._FAKE_MNT_BASE

        self.context = context.get_admin_context()
        vol_type = fake_volume.fake_volume_type_obj(self.context)
        vol_type.extra_specs = {}
        _FAKE_VOLUME = {'id': '4f711859-4928-4cb7-801a-a50c37ceaccc',
                        'size': 1,
                        'provider_location': self._FAKE_SHARE,
                        'name': self._FAKE_VOLUME_NAME,
                        'status': 'available'}
        self.vol = fake_volume.fake_volume_obj(self.context,
                                               volume_type_id=vol_type.id,
                                               **_FAKE_VOLUME)
        self.vol.volume_type = vol_type

        _FAKE_SNAPSHOT = {'id': self._FAKE_SNAPSHOT_ID,
                          'status': 'available',
                          'volume_size': 1}
        self.snap = fake_snapshot.fake_snapshot_obj(self.context,
                                                    **_FAKE_SNAPSHOT)
        self.snap.volume = self.vol

    def _path_exists(self, path):
        if path.startswith(self._cfg.vzstorage_shares_config):
            return True
        return _orig_path_exists(path)

    def _path_dont_exists(self, path):
        if path.startswith('/fake'):
            return False
        return _orig_path_exists(path)

    @mock.patch('os.path.exists')
    def test_setup_ok(self, mock_exists):
        mock_exists.side_effect = self._path_exists
        self._vz_driver.do_setup(mock.sentinel.context)

    @mock.patch('os.path.exists')
    def test_setup_missing_shares_conf(self, mock_exists):
        mock_exists.side_effect = self._path_dont_exists
        self.assertRaises(vzstorage.VzStorageException,
                          self._vz_driver.do_setup,
                          mock.sentinel.context)

    @mock.patch('os.path.exists')
    def test_setup_invalid_usage_ratio(self, mock_exists):
        mock_exists.side_effect = self._path_exists
        self._vz_driver.configuration.vzstorage_used_ratio = 1.2
        self.assertRaises(vzstorage.VzStorageException,
                          self._vz_driver.do_setup,
                          mock.sentinel.context)

    @mock.patch('os.path.exists')
    def test_setup_invalid_usage_ratio2(self, mock_exists):
        mock_exists.side_effect = self._path_exists
        self._vz_driver.configuration.vzstorage_used_ratio = 0
        self.assertRaises(vzstorage.VzStorageException,
                          self._vz_driver.do_setup,
                          mock.sentinel.context)

    @mock.patch('os.path.exists')
    def test_setup_invalid_mount_point_base(self, mock_exists):
        mock_exists.side_effect = self._path_exists
        self._cfg.vzstorage_mount_point_base = './tmp'
        vz_driver = vzstorage.VZStorageDriver(configuration=self._cfg)
        self.assertRaises(vzstorage.VzStorageException,
                          vz_driver.do_setup,
                          mock.sentinel.context)

    @mock.patch('os.path.exists')
    def test_setup_no_vzstorage(self, mock_exists):
        mock_exists.side_effect = self._path_exists
        exc = OSError()
        exc.errno = errno.ENOENT
        self._vz_driver._execute.side_effect = exc
        self.assertRaises(vzstorage.VzStorageException,
                          self._vz_driver.do_setup,
                          mock.sentinel.context)

    @ddt.data({'qemu_fmt': 'parallels', 'glance_fmt': 'ploop'},
              {'qemu_fmt': 'qcow2', 'glance_fmt': 'qcow2'})
    @ddt.unpack
    def test_initialize_connection(self, qemu_fmt, glance_fmt):
        drv = self._vz_driver
        info = mock.Mock()
        info.file_format = qemu_fmt
        snap_info = """{"volume_format": "%s",
                        "active": "%s"}""" % (qemu_fmt, self.vol.id)
        with mock.patch.object(drv, '_qemu_img_info', return_value=info):
            with mock.patch.object(drv, '_read_file',
                                   return_value=snap_info):
                ret = drv.initialize_connection(self.vol, None)
        name = drv.get_active_image_from_info(self.vol)
        expected = {'driver_volume_type': 'vzstorage',
                    'data': {'export': self._FAKE_SHARE,
                             'format': glance_fmt,
                             'name': name},
                    'mount_point_base': self._FAKE_MNT_BASE}
        self.assertEqual(expected, ret)

    def test_ensure_share_mounted_invalid_share(self):
        self.assertRaises(vzstorage.VzStorageException,
                          self._vz_driver._ensure_share_mounted, ':')

    @mock.patch.object(remotefs.RemoteFsClient, 'mount')
    def test_ensure_share_mounted(self, mock_mount):
        drv = self._vz_driver
        share = 'test'
        expected_calls = [
            mock.call(share, ['-u', 'cinder', '-g', 'root', '-l',
                              '/var/log/vstorage/%s/cinder.log.gz' % share]),
            mock.call(share, ['-l', '/var/log/dummy.log'])
        ]

        share_flags = '["-u", "cinder", "-g", "root"]'
        drv.shares[share] = share_flags
        drv._ensure_share_mounted(share)

        share_flags = '["-l", "/var/log/dummy.log"]'
        drv.shares[share] = share_flags
        drv._ensure_share_mounted(share)

        mock_mount.assert_has_calls(expected_calls)

    def test_find_share(self):
        drv = self._vz_driver
        drv._mounted_shares = [self._FAKE_SHARE]
        with mock.patch.object(drv, '_is_share_eligible', return_value=True):
            ret = drv._find_share(self.vol)
            self.assertEqual(self._FAKE_SHARE, ret)

    def test_find_share_no_shares_mounted(self):
        drv = self._vz_driver
        with mock.patch.object(drv, '_is_share_eligible', return_value=True):
            self.assertRaises(vzstorage.VzStorageNoSharesMounted,
                              drv._find_share, self.vol)

    def test_find_share_no_shares_suitable(self):
        drv = self._vz_driver
        drv._mounted_shares = [self._FAKE_SHARE]
        with mock.patch.object(drv, '_is_share_eligible', return_value=False):
            self.assertRaises(vzstorage.VzStorageNoSuitableShareFound,
                              drv._find_share, self.vol)

    def test_is_share_eligible_false(self):
        drv = self._vz_driver
        cap_info = (100 * units.Gi, 40 * units.Gi, 60 * units.Gi)
        with mock.patch.object(drv, '_get_capacity_info',
                               return_value=cap_info):
            ret = drv._is_share_eligible(self._FAKE_SHARE, 50)
            self.assertFalse(ret)

    def test_is_share_eligible_true(self):
        drv = self._vz_driver
        cap_info = (100 * units.Gi, 40 * units.Gi, 60 * units.Gi)
        with mock.patch.object(drv, '_get_capacity_info',
                               return_value=cap_info):
            ret = drv._is_share_eligible(self._FAKE_SHARE, 30)
            self.assertTrue(ret)

    @mock.patch.object(image_utils, 'resize_image')
    def test_extend_volume(self, mock_resize_image):
        drv = self._vz_driver
        drv._check_extend_volume_support = mock.Mock(return_value=True)
        drv._is_file_size_equal = mock.Mock(return_value=True)

        snap_info = '{"active": "%s"}' % self.vol.id
        with mock.patch.object(drv, 'get_volume_format',
                               return_value="raw"):
            with mock.patch.object(drv, 'get_active_image_from_info',
                                   return_value=self._FAKE_VOLUME_PATH):
                with mock.patch.object(drv, '_read_file',
                                       return_value=snap_info):
                    drv.extend_volume(self.vol, 10)

        mock_resize_image.assert_called_once_with(self._FAKE_VOLUME_PATH, 10)

    def _test_check_extend_support(self, is_eligible=True):
        drv = self._vz_driver
        drv.local_path = mock.Mock(return_value=self._FAKE_VOLUME_PATH)
        drv._is_share_eligible = mock.Mock(return_value=is_eligible)

        active = self._FAKE_VOLUME_PATH

        drv.get_active_image_from_info = mock.Mock(return_value=active)
        if not is_eligible:
            self.assertRaises(exception.ExtendVolumeError,
                              drv._check_extend_volume_support,
                              self.vol, 2)
        else:
            drv._check_extend_volume_support(self.vol, 2)
            drv._is_share_eligible.assert_called_once_with(self._FAKE_SHARE, 1)

    def test_check_extend_support(self):
        self._test_check_extend_support()

    def test_check_extend_volume_uneligible_share(self):
        self._test_check_extend_support(is_eligible=False)

    @mock.patch.object(image_utils, 'convert_image')
    def test_copy_volume_from_snapshot(self, mock_convert_image):
        drv = self._vz_driver

        fake_volume_info = {self._FAKE_SNAPSHOT_ID: 'fake_snapshot_file_name',
                            'backing-files':
                            {self._FAKE_SNAPSHOT_ID:
                             self._FAKE_VOLUME_NAME}}
        fake_img_info = mock.MagicMock()
        fake_img_info.backing_file = self._FAKE_VOLUME_NAME

        drv.get_volume_format = mock.Mock(return_value='raw')
        drv._local_path_volume_info = mock.Mock(
            return_value=self._FAKE_VOLUME_PATH + '.info')
        drv._local_volume_dir = mock.Mock(
            return_value=self._FAKE_MNT_POINT)
        drv._read_info_file = mock.Mock(
            return_value=fake_volume_info)
        drv._qemu_img_info = mock.Mock(
            return_value=fake_img_info)
        drv.local_path = mock.Mock(
            return_value=self._FAKE_VOLUME_PATH[:-1])
        drv._extend_volume = mock.Mock()

        drv._copy_volume_from_snapshot(
            self.snap, self.vol,
            self.vol['size'])
        drv._extend_volume.assert_called_once_with(
            self.vol, self.vol['size'], 'raw')
        mock_convert_image.assert_called_once_with(
            self._FAKE_VOLUME_PATH, self._FAKE_VOLUME_PATH[:-1], 'raw')

    def test_delete_volume(self):
        drv = self._vz_driver
        fake_vol_info = self._FAKE_VOLUME_PATH + '.info'

        drv._ensure_share_mounted = mock.MagicMock()
        fake_ensure_mounted = drv._ensure_share_mounted

        drv._local_volume_dir = mock.Mock(
            return_value=self._FAKE_MNT_POINT)
        drv.get_active_image_from_info = mock.Mock(
            return_value=self._FAKE_VOLUME_NAME)
        drv._delete = mock.Mock()
        drv._local_path_volume_info = mock.Mock(
            return_value=fake_vol_info)

        with mock.patch('os.path.exists', lambda x: True):
            drv.delete_volume(self.vol)

            fake_ensure_mounted.assert_called_once_with(self._FAKE_SHARE)
            drv._delete.assert_any_call(
                self._FAKE_VOLUME_PATH)
            drv._delete.assert_any_call(fake_vol_info)

    @mock.patch('cinder.volume.drivers.remotefs.RemoteFSSnapDriverBase.'
                '_write_info_file')
    def test_delete_snapshot_ploop(self, _mock_write_info_file):
        fake_snap_info = {
            'active': self._FAKE_VOLUME_NAME,
            self._FAKE_SNAPSHOT_ID: self._FAKE_SNAPSHOT_PATH,
        }
        self._vz_driver.get_volume_format = mock.Mock(
            return_value=vzstorage.DISK_FORMAT_PLOOP)
        self._vz_driver._read_info_file = mock.Mock(
            return_value=fake_snap_info
        )
        self._vz_driver._get_desc_path = mock.Mock(
            return_value='%s/DiskDescriptor.xml' % self._FAKE_VOLUME_PATH
        )
        self._vz_driver.delete_snapshot(self.snap)
        self._vz_driver._execute.assert_called_once_with(
            'ploop', 'snapshot-delete', '-u',
            '{%s}' % self._FAKE_SNAPSHOT_ID,
            '%s/DiskDescriptor.xml' % self._FAKE_VOLUME_PATH,
            run_as_root=True
        )

    @mock.patch('cinder.volume.drivers.remotefs.RemoteFSSnapDriverBase.'
                '_delete_snapshot')
    def test_delete_snapshot_qcow2_invalid_snap_info(self,
                                                     mock_delete_snapshot):
        fake_snap_info = {
            'active': self._FAKE_VOLUME_NAME,
        }
        self._vz_driver.get_volume_format = mock.Mock(
            return_value=vzstorage.DISK_FORMAT_QCOW2)
        self._vz_driver._read_info_file = mock.Mock(
            return_value=fake_snap_info
        )
        self._vz_driver.delete_snapshot(self.snap)
        self.assertFalse(mock_delete_snapshot.called)

    def test_extend_volume_ploop(self):
        drv = self._vz_driver
        drv.get_active_image_from_info = mock.Mock(
            return_value=self._FAKE_VOLUME_PATH)
        drv.get_volume_format = mock.Mock(
            return_value=vzstorage.DISK_FORMAT_PLOOP)
        drv._is_share_eligible = mock.Mock(
            return_value=True)
        drv.extend_volume(self.vol, 100)
        drv._execute.assert_called_once_with(
            'ploop', 'resize', '-s', '100G',
            '%s/DiskDescriptor.xml' % self._FAKE_VOLUME_PATH,
            run_as_root=True)

    @mock.patch.object(os.path, 'exists', return_value=False)
    def test_do_create_volume_with_volume_type(self, mock_exists):
        drv = self._vz_driver
        drv.local_path = mock.Mock(
            return_value=self._FAKE_VOLUME_PATH)
        drv._write_info_file = mock.Mock()
        drv._qemu_img_info = mock.Mock()
        drv._create_qcow2_file = mock.Mock()
        drv._create_ploop = mock.Mock()

        volume_type = fake_volume.fake_volume_type_obj(self.context)
        volume_type.extra_specs = {
            'vz:volume_format': 'qcow2'
        }
        volume1 = fake_volume.fake_volume_obj(self.context)
        volume1.size = 1024
        volume1.volume_type = volume_type
        volume2 = copy.deepcopy(volume1)
        volume2.metadata = {
            'volume_format': 'ploop'
        }

        drv._do_create_volume(volume1)
        drv._create_qcow2_file.assert_called_once_with(
            self._FAKE_VOLUME_PATH, 1024)

        drv._do_create_volume(volume2)
        drv._create_ploop.assert_called_once_with(
            self._FAKE_VOLUME_PATH, 1024)

    @mock.patch('cinder.volume.drivers.remotefs.RemoteFSSnapDriver.'
                '_create_cloned_volume')
    @mock.patch.object(vzstorage.VZStorageDriver, 'get_volume_format',
                       return_value='qcow2')
    def test_create_cloned_volume_qcow2(self,
                                        mock_get_volume_format,
                                        mock_remotefs_create_cloned_volume,
                                        ):
        drv = self._vz_driver
        volume = fake_volume.fake_volume_obj(self.context)
        src_vref_id = '375e32b2-804a-49f2-b282-85d1d5a5b9e1'
        src_vref = fake_volume.fake_volume_obj(
            self.context,
            id=src_vref_id,
            name='volume-%s' % src_vref_id,
            provider_location=self._FAKE_SHARE)
        src_vref.context = self.context

        mock_remotefs_create_cloned_volume.return_value = {
            'provider_location': self._FAKE_SHARE}
        ret = drv.create_cloned_volume(volume, src_vref)
        # Bug #1875953: code should also be passing context here
        mock_remotefs_create_cloned_volume.assert_called_once_with(
            volume, src_vref, self.context)
        self.assertEqual(ret, {'provider_location': self._FAKE_SHARE})

    @mock.patch.object(vzstorage.VZStorageDriver, '_local_path_volume_info')
    @mock.patch.object(vzstorage.VZStorageDriver, '_create_snapshot_ploop')
    @mock.patch.object(vzstorage.VZStorageDriver, 'delete_snapshot')
    @mock.patch.object(vzstorage.VZStorageDriver, '_write_info_file')
    @mock.patch.object(vzstorage.VZStorageDriver, '_copy_volume_from_snapshot')
    @mock.patch.object(vzstorage.VZStorageDriver, 'get_volume_format',
                       return_value='ploop')
    def test_create_cloned_volume_ploop(self,
                                        mock_get_volume_format,
                                        mock_copy_volume_from_snapshot,
                                        mock_write_info_file,
                                        mock_delete_snapshot,
                                        mock_create_snapshot_ploop,
                                        mock_local_path_volume_info,
                                        ):
        drv = self._vz_driver
        volume = fake_volume.fake_volume_obj(self.context)
        src_vref_id = '375e32b2-804a-49f2-b282-85d1d5a5b9e1'
        src_vref = fake_volume.fake_volume_obj(
            self.context,
            id=src_vref_id,
            name='volume-%s' % src_vref_id,
            provider_location=self._FAKE_SHARE)
        src_vref.context = self.context

        snap_attrs = ['volume_name', 'size', 'volume_size', 'name',
                      'volume_id', 'id', 'volume']
        Snapshot = collections.namedtuple('Snapshot', snap_attrs)

        snap_ref = Snapshot(volume_name=volume.name,
                            name='clone-snap-%s' % src_vref.id,
                            size=src_vref.size,
                            volume_size=src_vref.size,
                            volume_id=src_vref.id,
                            id=src_vref.id,
                            volume=src_vref)

        def _check_provider_location(volume):
            self.assertEqual(volume.provider_location, self._FAKE_SHARE)
            return mock.sentinel.fake_info_path
        mock_local_path_volume_info.side_effect = _check_provider_location

        ret = drv.create_cloned_volume(volume, src_vref)
        self.assertEqual(ret, {'provider_location': self._FAKE_SHARE})

        mock_write_info_file.assert_called_once_with(
            mock.sentinel.fake_info_path, {'active': 'volume-%s' % volume.id})
        mock_create_snapshot_ploop.assert_called_once_with(snap_ref)
        mock_copy_volume_from_snapshot.assert_called_once_with(
            snap_ref, volume, volume.size)
        mock_delete_snapshot.assert_called_once_with(snap_ref)



# Copyright (c) 2016 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for cluster table related operations."""

from unittest import mock

from sqlalchemy.orm import exc

from cinder import db
from cinder import exception
from cinder.tests.unit import test_db_api
from cinder.tests.unit import utils


class ClusterTestCase(test_db_api.BaseTest):
    """Unit tests for cinder.db.api.cluster_*."""

    def test_cluster_create_and_get(self):
        """Basic cluster creation test."""
        values = utils.default_cluster_values()
        cluster = db.cluster_create(self.ctxt, values)
        values['last_heartbeat'] = None
        self.assertEqual(0, cluster.race_preventer)
        for k, v in values.items():
            self.assertEqual(v, getattr(cluster, k))

        db_cluster = db.cluster_get(self.ctxt, cluster.id,
                                    services_summary=True)
        for k, v in values.items():
            self.assertEqual(v, getattr(db_cluster, k))
        self.assertEqual(0, db_cluster.race_preventer)

    def test_cluster_create_cfg_disabled(self):
        """Test that create uses enable_new_services configuration option."""
        self.override_config('enable_new_services', False)
        cluster = utils.create_cluster(self.ctxt, disabled=None)
        self.assertTrue(cluster.disabled)

    def test_cluster_create_disabled_preference(self):
        """Test that provided disabled value has highest priority on create."""
        self.override_config('enable_new_services', False)
        cluster = utils.create_cluster(self.ctxt)
        self.assertFalse(cluster.disabled)

    def test_cluster_create_duplicate(self):
        """Test that unique constraints are working.

        To remove potential races on creation we have a constraint set on name
        and race_preventer fields, and we set value on creation to 0, so 2
        clusters with the same name will fail this constraint.  On deletion we
        change this field to the same value as the id which will be unique and
        will not conflict with the creation of another cluster with the same
        name.
        """
        cluster = utils.create_cluster(self.ctxt)
        self.assertRaises(exception.ClusterExists,
                          utils.create_cluster,
                          self.ctxt,
                          name=cluster.name)

    def test_cluster_create_not_duplicate(self):
        """Test that unique constraints will work with delete operation.

        To remove potential races on creation we have a constraint set on name
        and race_preventer fields, and we set value on creation to 0, so 2
        clusters with the same name will fail this constraint.  On deletion we
        change this field to the same value as the id which will be unique and
        will not conflict with the creation of another cluster with the same
        name.
        """
        cluster = utils.create_cluster(self.ctxt)
        self.assertIsNone(db.cluster_destroy(self.ctxt, cluster.id))
        self.assertIsNotNone(utils.create_cluster(self.ctxt,
                                                  name=cluster.name))

    def test_cluster_get_fail(self):
        """Test that cluster get will fail if the cluster doesn't exists."""
        utils.create_cluster(self.ctxt, name='cluster@backend')
        self.assertRaises(exception.ClusterNotFound,
                          db.cluster_get, self.ctxt, 'name=cluster@backend2')

    def test_cluster_get_by_name(self):
        """Getting a cluster by name will include backends if not specified."""
        cluster = utils.create_cluster(self.ctxt, name='cluster@backend')
        # Get without the backend
        db_cluster = db.cluster_get(self.ctxt, name='cluster')
        self.assertEqual(cluster.id, db_cluster.id)
        # Get with the backend detail
        db_cluster = db.cluster_get(self.ctxt, name='cluster@backend')
        self.assertEqual(cluster.id, db_cluster.id)

    def test_cluster_get_without_summary(self):
        """Test getting cluster without summary information."""
        cluster = utils.create_cluster(self.ctxt)
        db_cluster = db.cluster_get(self.ctxt, cluster.id)
        self.assertRaises(exc.DetachedInstanceError,
                          getattr, db_cluster, 'num_hosts')
        self.assertRaises(exc.DetachedInstanceError,
                          getattr, db_cluster, 'num_down_hosts')
        self.assertIsNone(db_cluster.last_heartbeat)

    def test_cluster_get_with_summary_empty_cluster(self):
        """Test getting empty cluster with summary information."""
        cluster = utils.create_cluster(self.ctxt)
        db_cluster = db.cluster_get(self.ctxt, cluster.id,
                                    services_summary=True)
        self.assertEqual(0, db_cluster.num_hosts)
        self.assertEqual(0, db_cluster.num_down_hosts)
        self.assertIsNone(db_cluster.last_heartbeat)

    def test_cluster_get_with_summary(self):
        """Test getting cluster with summary information."""
        cluster, svcs = utils.create_populated_cluster(self.ctxt, 3, 1)
        db_cluster = db.cluster_get(self.ctxt, cluster.id,
                                    services_summary=True)
        self.assertEqual(3, db_cluster.num_hosts)
        self.assertEqual(1, db_cluster.num_down_hosts)
        self.assertEqual(svcs[1].updated_at, db_cluster.last_heartbeat)

    def test_cluster_get_is_up_on_empty_cluster(self):
        """Test is_up filter works on empty clusters."""
        cluster = utils.create_cluster(self.ctxt)
        db_cluster = db.cluster_get(self.ctxt, cluster.id, is_up=False)
        self.assertEqual(cluster.id, db_cluster.id)
        self.assertRaises(exception.ClusterNotFound,
                          db.cluster_get, self.ctxt, cluster.id, is_up=True)

    def test_cluster_get_services_on_empty_cluster(self):
        """Test get_services filter works on empty clusters."""
        cluster = utils.create_cluster(self.ctxt)
        db_cluster = db.cluster_get(self.ctxt, cluster.id, get_services=True)
        self.assertEqual(cluster.id, db_cluster.id)
        self.assertListEqual([], db_cluster.services)

    def test_cluster_get_services(self):
        """Test services is properly populated on non empty cluster."""
        # We create another cluster to see we do the selection correctly
        utils.create_populated_cluster(self.ctxt, 2, name='cluster2')
        # We create our cluster with 2 up nodes and 1 down
        cluster, svcs = utils.create_populated_cluster(self.ctxt, 3, 1)
        # Add a deleted service to the cluster
        db.service_create(self.ctxt,
                          {'cluster_name': cluster.name,
                           'deleted': True})
        db_cluster = db.cluster_get(self.ctxt, name=cluster.name,
                                    get_services=True)
        self.assertEqual(3, len(db_cluster.services))
        self.assertSetEqual({svc.id for svc in svcs},
                            {svc.id for svc in db_cluster.services})

    def test_cluster_get_is_up_all_are_down(self):
        """Test that is_up filter works when all services are down."""
        cluster, svcs = utils.create_populated_cluster(self.ctxt, 3, 3)
        self.assertRaises(exception.ClusterNotFound,
                          db.cluster_get, self.ctxt, cluster.id, is_up=True)
        db_cluster = db.cluster_get(self.ctxt, name=cluster.name, is_up=False)
        self.assertEqual(cluster.id, db_cluster.id)

    def test_cluster_get_by_num_down_hosts(self):
        """Test cluster_get by subquery field num_down_hosts."""
        cluster, svcs = utils.create_populated_cluster(self.ctxt, 3, 2)
        result = db.cluster_get(self.ctxt, num_down_hosts=2)
        self.assertEqual(cluster.id, result.id)

    def test_cluster_get_by_num_hosts(self):
        """Test cluster_get by subquery field num_hosts."""
        cluster, svcs = utils.create_populated_cluster(self.ctxt, 3, 2)
        result = db.cluster_get(self.ctxt, num_hosts=3)
        self.assertEqual(cluster.id, result.id)

    def test_cluster_destroy(self):
        """Test basic cluster destroy."""
        cluster = utils.create_cluster(self.ctxt)
        # On creation race_preventer is marked with a 0
        self.assertEqual(0, cluster.race_preventer)
        db.cluster_destroy(self.ctxt, cluster.id)
        db_cluster = db.cluster_get(self.ctxt, cluster.id, read_deleted='yes')
        self.assertTrue(db_cluster.deleted)
        self.assertIsNotNone(db_cluster.deleted_at)
        # On deletion race_preventer is marked with the id
        self.assertEqual(cluster.id, db_cluster.race_preventer)

    def test_cluster_destroy_non_existent(self):
        """Test destroying non existent cluster."""
        self.assertRaises(exception.ClusterNotFound,
                          db.cluster_destroy, self.ctxt, 0)

    def test_cluster_destroy_has_services(self):
        """Test that we cannot delete a cluster with non deleted services."""
        cluster, svcs = utils.create_populated_cluster(self.ctxt, 3, 1)
        self.assertRaises(exception.ClusterHasHosts,
                          db.cluster_destroy, self.ctxt, cluster.id)

    def test_cluster_update_non_existent(self):
        """Test that we raise an exception on updating non existent cluster."""
        self.assertRaises(exception.ClusterNotFound,
                          db.cluster_update, self.ctxt, 0, {'disabled': True})

    def test_cluster_update(self):
        """Test basic cluster update."""
        cluster = utils.create_cluster(self.ctxt)
        self.assertFalse(cluster.disabled)
        db.cluster_update(self.ctxt, cluster.id, {'disabled': True})
        db_cluster = db.cluster_get(self.ctxt, cluster.id)
        self.assertTrue(db_cluster.disabled)

    def test_cluster_get_all_empty(self):
        """Test basic empty cluster get_all."""
        self.assertListEqual([], db.cluster_get_all(self.ctxt))

    def test_cluster_get_all_matches(self):
        """Basic test of get_all with a matching filter."""
        cluster1, svcs = utils.create_populated_cluster(self.ctxt, 3, 1)
        cluster2, svcs = utils.create_populated_cluster(self.ctxt, 3, 2,
                                                        name='cluster2')
        cluster3, svcs = utils.create_populated_cluster(self.ctxt, 3, 3,
                                                        name='cluster3')

        expected = {cluster1.id, cluster2.id}
        result = db.cluster_get_all(self.ctxt, is_up=True)
        self.assertEqual(len(expected), len(result))
        self.assertSetEqual(expected, {cluster.id for cluster in result})

    def test_cluster_get_all_no_match(self):
        """Basic test of get_all with a non matching filter."""
        cluster1, svcs = utils.create_populated_cluster(self.ctxt, 3, 3)
        result = db.cluster_get_all(self.ctxt, is_up=True)
        self.assertListEqual([], result)

    @mock.patch('cinder.db.sqlalchemy.api._cluster_query')
    def test_cluster_get_all_passes_parameters(self, cluster_query_mock):
        """Test that get_all passes all parameters.

        Since we have already tested all filters and parameters with
        cluster_get method all we have to do for get_all is to check that we
        are passing them to the query building method.
        """
        args = (mock.sentinel.read_deleted, mock.sentinel.get_services,
                mock.sentinel.services_summary, mock.sentinel.is_up,
                mock.sentinel.name_match_level)
        filters = {'session': mock.sentinel.session,
                   'name': mock.sentinel.name,
                   'disabled': mock.sentinel.disabled,
                   'disabled_reason': mock.sentinel.disabled_reason,
                   'race_preventer': mock.sentinel.race_preventer,
                   'last_heartbeat': mock.sentinel.last_heartbeat,
                   'num_hosts': mock.sentinel.num_hosts,
                   'num_down_hosts': mock.sentinel.num_down_hosts}
        db.cluster_get_all(self.ctxt, *args, **filters)
        cluster_query_mock.assert_called_once_with(self.ctxt, *args, **filters)



#    (c)  Copyright  Kioxia Corporation 2021 All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Volume driver for KIOXIA KumoScale NVMeOF storage system."""


from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils.secretutils import md5

from cinder.common import constants
from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder.volume import driver
from cinder.volume.drivers.kioxia import entities
from cinder.volume.drivers.kioxia import rest_client

LOG = logging.getLogger(__name__)

KUMOSCALE_OPTS = [
    cfg.StrOpt("kioxia_url", help="KumoScale provisioner REST API URL"),
    cfg.StrOpt("kioxia_cafile", help="Cert for provisioner REST API SSL"),
    cfg.StrOpt("kioxia_token", help="KumoScale Provisioner auth token."),
    cfg.IntOpt(
        "kioxia_num_replicas", default=1,
        help="Number of volume replicas."),
    cfg.IntOpt(
        "kioxia_max_iops_per_gb", default=0, help="Upper limit for IOPS/GB."),
    cfg.IntOpt(
        "kioxia_desired_iops_per_gb", default=0, help="Desired IOPS/GB."),
    cfg.IntOpt(
        "kioxia_max_bw_per_gb", default=0,
        help="Upper limit for bandwidth in B/s per GB."),
    cfg.IntOpt(
        "kioxia_desired_bw_per_gb", default=0,
        help="Desired bandwidth in B/s per GB."),
    cfg.BoolOpt(
        "kioxia_same_rack_allowed", default=False,
        help="Can more than one replica be allocated to same rack."),
    cfg.IntOpt(
        "kioxia_block_size", default=4096,
        help="Volume block size in bytes - 512 or 4096 (Default)."),
    cfg.BoolOpt(
        "kioxia_writable", default=False,
        help="Volumes from snapshot writeable or not."),
    cfg.StrOpt(
        "kioxia_provisioning_type", default="THICK",
        choices=[
            ('THICK', 'Thick provisioning'), ('THIN', 'Thin provisioning')],
        help="Thin or thick volume, Default thick."),
    cfg.IntOpt(
        "kioxia_vol_reserved_space_percentage", default=0,
        help="Thin volume reserved capacity allocation percentage."),
    cfg.IntOpt(
        "kioxia_snap_reserved_space_percentage", default=0,
        help="Percentage of the parent volume to be used for log."),
    cfg.IntOpt(
        "kioxia_snap_vol_reserved_space_percentage", default=0,
        help="Writable snapshot percentage of parent volume used for log."),
    cfg.IntOpt(
        "kioxia_max_replica_down_time", default=0,
        help="Replicated volume max downtime for replica in minutes."),
    cfg.BoolOpt(
        "kioxia_span_allowed", default=True,
        help="Allow span - Default True."),
    cfg.BoolOpt(
        "kioxia_snap_vol_span_allowed", default=True,
        help="Allow span in snapshot volume - Default True.")
]

CONF = cfg.CONF
CONF.register_opts(KUMOSCALE_OPTS)


@interface.volumedriver
class KumoScaleBaseVolumeDriver(driver.BaseVD):
    """Performs volume management on KumoScale Provisioner.

    Version history:

    .. code-block:: none

        1.0.0 - Initial driver version.
    """

    VERSION = '1.0.0'
    CI_WIKI_NAME = 'KIOXIA_CI'
    SUPPORTED_REST_API_VERSIONS = ['1.0', '1.1']

    def __init__(self, *args, **kwargs):
        super(KumoScaleBaseVolumeDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(KUMOSCALE_OPTS)
        self._backend_name = (
            self.configuration.volume_backend_name or self.__class__.__name__)
        self.kumoscale = self._get_kumoscale(
            self.configuration.safe_get("kioxia_url"),
            self.configuration.safe_get("kioxia_token"),
            self.configuration.safe_get("kioxia_cafile"))

        self.num_replicas = self.configuration.safe_get("kioxia_num_replicas")
        self.same_rack_allowed = self.configuration.safe_get(
            "kioxia_same_rack_allowed")
        self.max_iops_per_gb = self.configuration.safe_get(
            "kioxia_max_iops_per_gb")
        self.desired_iops_per_gb = self.configuration.safe_get(
            "kioxia_desired_iops_per_gb")
        self.max_bw_per_gb = self.configuration.safe_get(
            "kioxia_max_bw_per_gb")
        self.desired_bw_per_gb = self.configuration.safe_get(
            "kioxia_desired_bw_per_gb")
        self.block_size = self.configuration.safe_get("kioxia_block_size")
        self.writable = self.configuration.safe_get("kioxia_writable")
        self.provisioning_type = self.configuration.safe_get(
            "kioxia_provisioning_type")
        self.vol_reserved_space_percentage = self.configuration.safe_get(
            "kioxia_vol_reserved_space_percentage")
        self.snap_vol_reserved_space_percentage = self.configuration.safe_get(
            "kioxia_snap_vol_reserved_space_percentage")
        self.snap_reserved_space_percentage = self.configuration.safe_get(
            "kioxia_snap_reserved_space_percentage")
        self.max_replica_down_time = self.configuration.safe_get(
            "kioxia_max_replica_down_time")
        self.span_allowed = self.configuration.safe_get("kioxia_span_allowed")
        self.snap_vol_span_allowed = self.configuration.safe_get(
            "kioxia_snap_vol_span_allowed")

    @staticmethod
    def get_driver_options():
        return KUMOSCALE_OPTS

    def _get_kumoscale(self, url, token, cert):
        """Returns an initialized rest client"""
        url_strs = url.split(":")
        ip_str = url_strs[1]
        ip_strs = ip_str.split("//")
        ip = ip_strs[1]
        port = url_strs[2]
        kumoscale = rest_client.KioxiaProvisioner([ip], cert, token, port)
        return kumoscale

    def create_volume(self, volume):
        """Create the volume"""
        volume_name = volume["name"]
        volume_uuid = volume["id"]
        volume_size = volume["size"]
        zone_list = None if 'availability_zone' not in volume else [
            volume['availability_zone']]

        if self.num_replicas > 1 and len(volume_name) > 27:
            volume_name = volume_name[:27]  # workaround for limitation
        storage_class = entities.StorageClass(
            self.num_replicas, None, None, zone_list, self.block_size,
            self.max_iops_per_gb, self.desired_iops_per_gb, self.max_bw_per_gb,
            self.desired_bw_per_gb, self.same_rack_allowed,
            self.max_replica_down_time, None, self.span_allowed)
        ks_volume = entities.VolumeCreate(
            volume_name, volume_size, storage_class, self.provisioning_type,
            self.vol_reserved_space_percentage, 'NVMeoF', volume_uuid)

        try:
            result = self.kumoscale.create_volume(ks_volume)
        except Exception as e:
            msg = (_("Volume %(volname)s creation exception: %(txt)s") %
                   {'volname': volume_name, 'txt': str(e)})
            raise exception.VolumeBackendAPIException(data=msg)

        if result.status != 'Success':
            raise exception.VolumeBackendAPIException(data=result.description)

    def delete_volume(self, volume):
        """Delete the volume"""
        volume_uuid = volume["id"]

        try:
            result = self.kumoscale.delete_volume(volume_uuid)
        except Exception as e:
            msg = (_("Volume %(voluuid)s deletion exception: %(txt)s") %
                   {'voluuid': volume_uuid, 'txt': str(e)})
            raise exception.VolumeBackendAPIException(data=msg)

        if result.status not in ('Success', 'DeviceNotFound', 'NotExists'):
            raise exception.VolumeBackendAPIException(data=result.description)

    def create_snapshot(self, snapshot):

        snapshot_name = snapshot['name']
        snapshot_uuid = snapshot['id']
        volume_uuid = snapshot['volume_id']
        ks_snapshot = entities.SnapshotCreate(
            snapshot_name, volume_uuid,
            self.snap_reserved_space_percentage, snapshot_uuid)

        try:
            result = self.kumoscale.create_snapshot(ks_snapshot)
        except Exception as e:
            msg = (_("Snapshot %(snapname)s creation exception: %(txt)s") %
                   {'snapname': snapshot_name, 'txt': str(e)})
            raise exception.VolumeBackendAPIException(data=msg)

        if result.status != 'Success':
            raise exception.VolumeBackendAPIException(data=result.description)

    def delete_snapshot(self, snapshot):

        snapshot_uuid = snapshot['id']

        try:
            result = self.kumoscale.delete_snapshot(snapshot_uuid)
        except Exception as e:
            msg = (_("Snapshot %(snapuuid)s deletion exception: %(txt)s") %
                   {'snapuuid': snapshot_uuid, 'txt': str(e)})
            raise exception.VolumeBackendAPIException(data=msg)

        if result.status not in ('Success', 'DeviceNotFound', 'NotExists'):
            raise exception.VolumeBackendAPIException(data=result.description)

    def create_volume_from_snapshot(self, volume, snapshot):

        volume_name = volume["name"]
        volume_uuid = volume["id"]
        snapshot_uuid = snapshot["id"]
        if self.writable:
            reserved_space_percentage = self.snap_vol_reserved_space_percentage
        else:
            reserved_space_percentage = 0

        ks_snapshot_volume = entities.SnapshotVolumeCreate(
            volume_name, snapshot_uuid, self.writable,
            reserved_space_percentage, volume_uuid,
            self.max_iops_per_gb, self.max_bw_per_gb, 'NVMeoF',
            self.snap_vol_span_allowed)

        try:
            result = self.kumoscale.create_snapshot_volume(ks_snapshot_volume)
        except Exception as e:
            msg = (_("Volume %(volname)s from snapshot exception: %(txt)s") %
                   {'volname': volume_name, 'txt': str(e)})
            raise exception.VolumeBackendAPIException(data=msg)

        if result.status != 'Success':
            raise exception.VolumeBackendAPIException(data=result.description)

    def initialize_connection(self, volume, connector, initiator_data=None):
        """Connect the initiator to a volume"""
        host_uuid = connector['uuid']
        ks_volume = None
        targets = []
        volume_replicas = []
        volume_uuid = volume['id']
        volume_name = volume['name']

        try:
            result = self.kumoscale.host_probe(
                connector['nqn'], connector['uuid'],
                KumoScaleBaseVolumeDriver._convert_host_name(
                    connector['host']),
                'Agent', 'cinder-driver-0.1', 30)
        except Exception as e:
            msg = (_("Host %(uuid)s host_probe exception: %(txt)s") %
                   {'uuid': connector['uuid'], 'txt': str(e)})
            raise exception.VolumeBackendAPIException(data=msg)

        if result.status != 'Success':
            msg = (_("host_probe for %(uuid)s failed with %(txt)s") %
                   {'uuid': connector['uuid'], 'txt': result.description})
            raise exception.VolumeBackendAPIException(data=msg)

        try:
            result = self.kumoscale.publish(host_uuid, volume_uuid)
        except Exception as e:
            msg = (_("Volume %(voluuid)s publish exception: %(txt)s") %
                   {'voluuid': volume_uuid, 'txt': str(e)})
            raise exception.VolumeBackendAPIException(data=msg)

        if result.status != "Success" and result.status != 'AlreadyPublished':
            raise exception.VolumeBackendAPIException(data=result.description)

        try:
            result = self.kumoscale.get_volumes_by_uuid(volume_uuid)
        except Exception as e:
            msg = (_("Volume %(voluuid)s fetch exception: %(txt)s") %
                   {'voluuid': volume_uuid, 'txt': str(e)})
            raise exception.VolumeBackendAPIException(data=msg)

        if result.status == "Success":
            if len(result.prov_entities) == 0:
                raise exception.VolumeBackendAPIException(
                    data=_("Volume %s not found") % volume_uuid)
            else:
                ks_volume = result.prov_entities[0]
        else:
            msg = (_("get_volumes_by_uuid for %(uuid)s failed with %(txt)s") %
                   {'uuid': volume_uuid, 'txt': result.description})
            raise exception.VolumeBackendAPIException(data=msg)

        try:
            result = self.kumoscale.get_targets(host_uuid, ks_volume.uuid)
        except Exception as e:
            msg = (_("Volume %(voluuid)s get targets exception: %(txt)s") %
                   {'voluuid': volume_uuid, 'txt': str(e)})
            raise exception.VolumeBackendAPIException(data=msg)

        if result.status == "Success":
            if len(result.prov_entities) == 0:
                raise exception.VolumeBackendAPIException(
                    data=_("Volume %s targets not found") % ks_volume.uuid)
            else:
                targets = result.prov_entities

        ks_volume_replicas = ks_volume.location
        for i in range(len(targets)):
            persistent_id = str(targets[i].backend.persistentID)

            try:
                result = self.kumoscale.get_backend_by_id(persistent_id)
            except Exception as e:
                msg = (_("Backend %(backpid)s exception: %(txt)s") %
                       {'backpid': persistent_id, 'txt': str(e)})
                raise exception.VolumeBackendAPIException(data=msg)

            if result.status == "Success":
                if len(result.prov_entities) == 0:
                    raise exception.VolumeBackendAPIException(
                        data=_("Backend %s not found") % persistent_id)
                else:
                    backend = result.prov_entities[0]
            else:
                msg = (_("get_backend_by_id for %(pid)s failed with %(txt)s") %
                       {'pid': persistent_id, 'txt': result.description})
                raise exception.VolumeBackendAPIException(data=msg)

            str_portals = []
            for p in range(len(backend.portals)):
                portal = backend.portals[p]
                portal_ip = str(portal.ip)
                portal_port = str(portal.port)
                portal_transport = str(portal.transport)
                str_portals.append(
                    (portal_ip, portal_port, portal_transport))

            for j in range(len(ks_volume_replicas)):
                ks_replica = ks_volume_replicas[j]
                if str(ks_replica.backend.persistentID) == persistent_id:
                    break

            replica = dict()
            replica['vol_uuid'] = ks_replica.uuid
            replica['target_nqn'] = str(targets[i].targetName)
            replica['portals'] = str_portals

            volume_replicas.append(replica)

        if len(volume_replicas) > 1:  # workaround for limitation
            volume_name = volume_name[:27]

        data = {
            'vol_uuid': volume_uuid,
            'alias': volume_name,
            'writable': ks_volume.writable,
            'volume_replicas': volume_replicas,
            'replica_count': len(ks_volume_replicas)
        }

        if result.status != 'Success':
            raise exception.VolumeBackendAPIException(data=result.description)

        return {
            'driver_volume_type': 'nvmeof',
            'data': data
        }

    @staticmethod
    def _convert_host_name(name):
        if name is None:
            return ""
        if len(name) > 32:
            name = md5(name.encode('utf-8'), usedforsecurity=False).hexdigest()
        else:
            name = name.replace('.', '-').lower()
        return name

    def terminate_connection(self, volume, connector, **kwargs):
        """Terminate connection."""
        volume_uuid = volume['id']
        if connector:
            host_uuid = connector['uuid']
        else:
            host_uuid = None

        try:
            result = self.kumoscale.unpublish(host_uuid, volume_uuid)
        except Exception as e:
            msg = (_("Volume %(voluuid)s unpublish exception: %(txt)s") %
                   {'voluuid': volume_uuid, 'txt': str(e)})
            raise exception.VolumeBackendAPIException(data=msg)

        if result.status != 'Success' and (
                result.status != 'VolumeNotPublished'):
            raise exception.VolumeBackendAPIException(data=result.description)

    def _update_volume_stats(self):
        data = dict(
            volume_backend_name=self._backend_name,
            vendor_name='KIOXIA',
            driver_version=self.VERSION,
            storage_protocol=constants.NVMEOF_VARIANT_1,
        )
        data['total_capacity_gb'] = 'unknown'
        data['free_capacity_gb'] = 'unknown'
        data['consistencygroup_support'] = False
        data['thin_provisioning_support'] = True
        data['multiattach'] = False

        result = None
        tenants = []
        try:
            result = self.kumoscale.get_tenants()
        except Exception as e:
            msg = _("Get tenants exception: %s") % str(e)
            LOG.exception(msg)

        if result and result.status == "Success":
            if len(result.prov_entities) == 0:
                LOG.error("No kumoscale tenants")
            else:
                tenants = result.prov_entities
        elif result:
            LOG.error("Get tenants API error: %s", result.description)

        default_tenant = None
        for i in range(len(tenants)):
            if tenants[i].tenantId == "0":
                default_tenant = tenants[i]
                break

        if default_tenant:
            total_capacity = default_tenant.capacity
            consumed_capacity = default_tenant.consumedCapacity
            free_capacity = total_capacity - consumed_capacity
            data['total_capacity_gb'] = total_capacity
            data['free_capacity_gb'] = free_capacity

        self._stats = data

    def extend_volume(self, volume, new_size):
        try:
            result = self.kumoscale.expand_volume(
                new_size, volume["id"])
        except Exception as e:
            msg = (_("Volume %(volid)s expand exception: %(txt)s") %
                   {'volid': volume["id"], 'txt': str(e)})
            raise exception.VolumeBackendAPIException(data=msg)
        if result.status != 'Success':
            raise exception.VolumeBackendAPIException(data=result.description)

    def create_cloned_volume(self, volume, src_vref):
        clone_entity = entities.CloneEntity(
            src_vref['id'], volume['name'],
            volumeId=volume['id'],
            capacity=volume['size'])
        try:
            result = self.kumoscale.clone_volume(clone_entity)
        except Exception as e:
            msg = (_("Volume %(volid)s clone exception: %(txt)s") %
                   {'volid': volume["id"], 'txt': str(e)})
            raise exception.VolumeBackendAPIException(data=msg)
        if result.status != 'Success':
            raise exception.VolumeBackendAPIException(data=result.description)

    def create_export(self, context, volume, connector):
        pass

    def ensure_export(self, context, volume):
        pass

    def remove_export(self, context, volume):
        pass

    def check_for_setup_error(self):
        pass



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Common VNC test needs."""

from cinder import context
from cinder.tests.unit import test
from cinder.volume import configuration as conf
from cinder.volume.drivers.dell_emc.vnx import common


class TestCase(test.TestCase):

    def setUp(self):
        super(TestCase, self).setUp()
        self.configuration = conf.Configuration(None)
        self.configuration.san_ip = '192.168.1.1'
        self.configuration.storage_vnx_authentication_type = 'global'
        self.configuration.config_group = 'vnx_backend'
        self.ctxt = context.get_admin_context()
        common.DEFAULT_TIMEOUT = 0
        common.INTERVAL_30_SEC = 0



# Copyright (c) 2012 NetApp, Inc.  All rights reserved.
# Copyright (c) 2014 Navneet Singh.  All rights reserved.
# Copyright (c) 2014 Bob Callaway.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Contains configuration options for NetApp drivers.

Common place to hold configuration options for all NetApp drivers.
Options need to be grouped into granular units to be able to be reused
by different modules and classes. This does not restrict declaring options in
individual modules. If options are not re usable then can be declared in
individual modules. It is recommended to Keep options at a single
place to ensure re usability and better management of configuration options.
"""

from oslo_config import cfg
from oslo_config import types

from cinder.volume import configuration as conf

NETAPP_SIZE_MULTIPLIER_DEFAULT = 1.2

netapp_proxy_opts = [
    cfg.StrOpt('netapp_storage_family',
               default='ontap_cluster',
               choices=['ontap_cluster'],
               help=('The storage family type used on the storage system; '
                     'the only valid value is ontap_cluster for using '
                     'clustered Data ONTAP.')),
    cfg.StrOpt('netapp_storage_protocol',
               choices=['iscsi', 'fc', 'nfs'],
               help=('The storage protocol to be used on the data path with '
                     'the storage system.')), ]

netapp_connection_opts = [
    cfg.StrOpt('netapp_server_hostname',
               help='The hostname (or IP address) for the storage system or '
                    'proxy server.'),
    cfg.IntOpt('netapp_server_port',
               help=('The TCP port to use for communication with the storage '
                     'system or proxy server. If not specified, Data ONTAP '
                     'drivers will use 80 for HTTP and 443 for HTTPS.')),
    cfg.BoolOpt('netapp_use_legacy_client',
                default=True,
                help=('Select which ONTAP client to use for retrieving and '
                      'modifying data on the storage. The legacy client '
                      'relies on ZAPI calls. If set to False, the new REST '
                      'client is used, which runs REST calls if supported, '
                      'otherwise falls back to the equivalent ZAPI call.')),
    cfg.IntOpt('netapp_async_rest_timeout',
               min=60,
               default=60,  # One minute
               help='The maximum time in seconds to wait for completing a '
                    'REST asynchronous operation.'), ]

netapp_transport_opts = [
    cfg.StrOpt('netapp_transport_type',
               default='http',
               choices=['http', 'https'],
               help=('The transport protocol used when communicating with '
                     'the storage system or proxy server.')),
    cfg.StrOpt('netapp_ssl_cert_path',
               help=("The path to a CA_BUNDLE file or directory with "
                     "certificates of trusted CA. If set to a directory, it "
                     "must have been processed using the c_rehash utility "
                     "supplied with OpenSSL. If not informed, it will use the "
                     "Mozilla's carefully curated collection of Root "
                     "Certificates for validating the trustworthiness of SSL "
                     "certificates. Only applies with new REST client.")), ]

netapp_basicauth_opts = [
    cfg.StrOpt('netapp_login',
               help=('Administrative user account name used to access the '
                     'storage system or proxy server.')),
    cfg.StrOpt('netapp_password',
               help=('Password for the administrative user account '
                     'specified in the netapp_login option.'),
               secret=True), ]

netapp_provisioning_opts = [
    cfg.FloatOpt('netapp_size_multiplier',
                 default=NETAPP_SIZE_MULTIPLIER_DEFAULT,
                 help=('The quantity to be multiplied by the requested '
                       'volume size to ensure enough space is available on '
                       'the virtual storage server (Vserver) to fulfill '
                       'the volume creation request.  Note: this option '
                       'is deprecated and will be removed in favor of '
                       '"reserved_percentage" in the Mitaka release.')),
    cfg.StrOpt('netapp_lun_space_reservation',
               default='enabled',
               choices=['enabled', 'disabled'],
               help=('This option determines if storage space is reserved '
                     'for LUN allocation. If enabled, LUNs are thick '
                     'provisioned. If space reservation is disabled, '
                     'storage space is allocated on demand.')),
    cfg.BoolOpt('netapp_driver_reports_provisioned_capacity',
                default=False,
                help=('Set to True for Cinder to query the storage system in '
                      'order to calculate volumes provisioned size, otherwise '
                      'provisioned_capacity_gb will corresponds to the '
                      'value of allocated_capacity_gb (calculated by Cinder '
                      'Core code). Enabling this feature increases '
                      'the number of API calls to the storage and '
                      'requires more processing on host, which may impact '
                      'volume report overall performance.')), ]

netapp_cluster_opts = [
    cfg.StrOpt('netapp_vserver',
               help=('This option specifies the virtual storage server '
                     '(Vserver) name on the storage cluster on which '
                     'provisioning of block storage volumes should occur.')), ]

netapp_img_cache_opts = [
    cfg.IntOpt('thres_avl_size_perc_start',
               default=20,
               help=('If the percentage of available space for an NFS share '
                     'has dropped below the value specified by this option, '
                     'the NFS image cache will be cleaned.')),
    cfg.IntOpt('thres_avl_size_perc_stop',
               default=60,
               help=('When the percentage of available space on an NFS share '
                     'has reached the percentage specified by this option, '
                     'the driver will stop clearing files from the NFS image '
                     'cache that have not been accessed in the last M '
                     'minutes, where M is the value of the '
                     'expiry_thres_minutes configuration option.')),
    cfg.IntOpt('expiry_thres_minutes',
               default=720,
               help=('This option specifies the threshold for last access '
                     'time for images in the NFS image cache. When a cache '
                     'cleaning cycle begins, images in the cache that have '
                     'not been accessed in the last M minutes, where M is '
                     'the value of this parameter, will be deleted from the '
                     'cache to create free space on the NFS share.')), ]

netapp_nfs_extra_opts = [
    cfg.StrOpt('netapp_copyoffload_tool_path',
               help=('This option specifies the path of the NetApp copy '
                     'offload tool binary. Ensure that the binary has execute '
                     'permissions set which allow the effective user of the '
                     'cinder-volume process to execute the file.'),
               deprecated_for_removal=True,
               deprecated_reason='The CopyOfflload tool is no longer '
                                 'available for downloading.'), ]
netapp_san_opts = [
    cfg.StrOpt('netapp_lun_ostype',
               help=('This option defines the type of operating system that'
                     ' will access a LUN exported from Data ONTAP; it is'
                     ' assigned to the LUN at the time it is created.')),
    cfg.StrOpt('netapp_host_type',
               help=('This option defines the type of operating system for'
                     ' all initiators that can access a LUN. This information'
                     ' is used when mapping LUNs to individual hosts or'
                     ' groups of hosts.')),
    cfg.StrOpt('netapp_pool_name_search_pattern',
               deprecated_opts=[cfg.DeprecatedOpt(name='netapp_volume_list'),
                                cfg.DeprecatedOpt(name='netapp_storage_pools')
                                ],
               default="(.+)",
               help=('This option is used to restrict provisioning to the '
                     'specified pools. Specify the value of '
                     'this option to be a regular expression which will be '
                     'applied to the names of objects from the storage '
                     'backend which represent pools in Cinder. This option '
                     'is only utilized when the storage protocol is '
                     'configured to use iSCSI or FC.')), ]

netapp_replication_opts = [
    cfg.MultiOpt('netapp_replication_aggregate_map',
                 item_type=types.Dict(),
                 help="Multi opt of dictionaries to represent the aggregate "
                      "mapping between source and destination back ends when "
                      "using whole back end replication. For every "
                      "source aggregate associated with a cinder pool (NetApp "
                      "FlexVol/FlexGroup), you would need to specify the "
                      "destination aggregate on the replication target "
                      "device. "
                      "A replication target device is configured with the "
                      "configuration option replication_device. Specify this "
                      "option as many times as you have replication devices. "
                      "Each entry takes the standard dict config form: "
                      "netapp_replication_aggregate_map = "
                      "backend_id:<name_of_replication_device_section>,"
                      "src_aggr_name1:dest_aggr_name1,"
                      "src_aggr_name2:dest_aggr_name2,..."),
    cfg.IntOpt('netapp_snapmirror_quiesce_timeout',
               min=0,
               default=3600,  # One Hour
               help='The maximum time in seconds to wait for existing '
                    'SnapMirror transfers to complete before aborting '
                    'during a failover.'),
    cfg.IntOpt('netapp_replication_volume_online_timeout',
               min=60,
               default=360,  # Default to six minutes
               help='Sets time in seconds to wait for a replication volume '
                    'create to complete and go online.')]

netapp_support_opts = [
    cfg.StrOpt('netapp_api_trace_pattern',
               default='(.*)',
               help=('A regular expression to limit the API tracing. This '
                     'option is honored only if enabling ``api`` tracing '
                     'with the ``trace_flags`` option. By default, '
                     'all APIs will be traced.')),
]

netapp_migration_opts = [
    cfg.IntOpt('netapp_migrate_volume_timeout',
               default=3600,
               min=30,
               help='Sets time in seconds to wait for storage assisted volume '
                    'migration to complete.'),
]

CONF = cfg.CONF
CONF.register_opts(netapp_proxy_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_connection_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_transport_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_basicauth_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_cluster_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_provisioning_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_img_cache_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_nfs_extra_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_san_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_replication_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_support_opts, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(netapp_migration_opts, group=conf.SHARED_CONF_GROUP)



#  Copyright 2014 Cloudbase Solutions Srl
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import copy
import os
import re
import sys
from unittest import mock

import castellan
import ddt

from cinder import context
from cinder import exception
from cinder.image import image_utils
from cinder.objects import fields
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit.keymgr import fake as fake_keymgr
from cinder.tests.unit import test
from cinder import utils
from cinder.volume.drivers import remotefs
from cinder.volume import volume_utils


class KeyObject(object):
    def get_encoded(arg):
        return "asdf".encode('utf-8')


@ddt.ddt
class RemoteFsSnapDriverTestCase(test.TestCase):

    _FAKE_MNT_POINT = '/mnt/fake_hash'

    def setUp(self):
        super(RemoteFsSnapDriverTestCase, self).setUp()
        self._driver = remotefs.RemoteFSSnapDriver()
        self._driver._remotefsclient = mock.Mock()
        self._driver._execute = mock.Mock()
        self._driver._delete = mock.Mock()

        self.context = context.get_admin_context()

        self._fake_volume = fake_volume.fake_volume_obj(
            self.context, provider_location='fake_share')
        self._fake_volume_path = os.path.join(self._FAKE_MNT_POINT,
                                              self._fake_volume.name)
        self._fake_snapshot = fake_snapshot.fake_snapshot_obj(self.context)
        self._fake_snapshot_path = (self._fake_volume_path + '.' +
                                    self._fake_snapshot.id)
        self._fake_snapshot.volume = self._fake_volume

        # Encrypted volume and snapshot
        self.volume_c = fake_volume.fake_volume_obj(
            self.context,
            **{'name': u'volume-0000000a',
               'id': '55555555-222f-4b32-b585-9991b3bf0a99',
               'size': 12,
               'encryption_key_id': fake.ENCRYPTION_KEY_ID})
        self._fake_snap_c = fake_snapshot.fake_snapshot_obj(self.context)
        self._fake_snap_c.volume = self.volume_c
        self.volume_c_path = os.path.join(self._FAKE_MNT_POINT,
                                          self.volume_c.name)
        self._fake_snap_c_path = (self.volume_c_path + '.' +
                                  self._fake_snap_c.id)

    @ddt.data({'current_state': 'in-use',
               'acceptable_states': ['available', 'in-use']},
              {'current_state': 'in-use',
               'acceptable_states': ['available'],
               'expected_exception': exception.InvalidVolume})
    @ddt.unpack
    def test_validate_state(self, current_state, acceptable_states,
                            expected_exception=None):
        if expected_exception:
            self.assertRaises(expected_exception,
                              self._driver._validate_state,
                              current_state,
                              acceptable_states)
        else:
            self._driver._validate_state(current_state, acceptable_states)

    def _test_delete_snapshot(self, volume_in_use=False,
                              stale_snapshot=False,
                              is_active_image=True,
                              is_tmp_snap=False,
                              encryption=False):
        # If the snapshot is not the active image, it is guaranteed that
        # another snapshot exists having it as backing file.

        fake_upper_snap_id = 'fake_upper_snap_id'
        if encryption:
            fake_snapshot_name = os.path.basename(self._fake_snap_c_path)
            fake_info = {'active': fake_snapshot_name,
                         self._fake_snap_c.id: fake_snapshot_name}
            expected_info = fake_info

            fake_upper_snap_path = (
                self.volume_c_path + '-snapshot' + fake_upper_snap_id)

            snapshot = self._fake_snap_c
            snapshot_path = self._fake_snap_c_path
            volume_name = self.volume_c.name
        else:
            fake_snapshot_name = os.path.basename(self._fake_snapshot_path)
            fake_info = {'active': fake_snapshot_name,
                         self._fake_snapshot.id: fake_snapshot_name}
            expected_info = fake_info

            fake_upper_snap_path = (
                self._fake_volume_path + '-snapshot' + fake_upper_snap_id)

            snapshot = self._fake_snapshot
            snapshot_path = self._fake_snapshot_path
            volume_name = self._fake_volume.name

        fake_snap_img_info = mock.Mock()
        fake_base_img_info = mock.Mock()
        if stale_snapshot:
            fake_snap_img_info.backing_file = None
        else:
            fake_snap_img_info.backing_file = volume_name
        fake_snap_img_info.file_format = 'qcow2'
        fake_base_img_info.backing_file = None
        fake_base_img_info.file_format = 'raw'

        self._driver._local_path_volume_info = mock.Mock(
            return_value=mock.sentinel.fake_info_path)
        self._driver._qemu_img_info = mock.Mock(
            side_effect=[fake_snap_img_info, fake_base_img_info])
        self._driver._local_volume_dir = mock.Mock(
            return_value=self._FAKE_MNT_POINT)

        self._driver._validate_state = mock.Mock()
        self._driver._read_info_file = mock.Mock()
        self._driver._write_info_file = mock.Mock()
        self._driver._img_commit = mock.Mock()
        self._driver._rebase_img = mock.Mock()
        self._driver._delete_stale_snapshot = mock.Mock()
        self._driver._delete_snapshot_online = mock.Mock()

        exp_acceptable_states = ['available', 'in-use', 'backing-up',
                                 'deleting', 'downloading']

        if volume_in_use:
            snapshot.volume.status = 'backing-up'
            snapshot.volume.attach_status = 'attached'

            self._driver._read_info_file.return_value = fake_info

            self._driver._delete_snapshot(snapshot)
            self._driver._validate_state.assert_called_once_with(
                snapshot.volume.status,
                exp_acceptable_states)
            if stale_snapshot:
                self._driver._delete_stale_snapshot.assert_called_once_with(
                    snapshot)
            else:
                expected_online_delete_info = {
                    'active_file': fake_snapshot_name,
                    'snapshot_file': fake_snapshot_name,
                    'base_file': volume_name,
                    'base_id': None,
                    'new_base_file': None
                }
                self._driver._delete_snapshot_online.assert_called_once_with(
                    self.context, snapshot,
                    expected_online_delete_info)

        elif is_active_image:
            self._driver._read_info_file.return_value = fake_info

            self._driver._delete_snapshot(snapshot)

            self._driver._img_commit.assert_called_once_with(
                snapshot_path)
            self.assertNotIn(snapshot.id, fake_info)
            self._driver._write_info_file.assert_called_once_with(
                mock.sentinel.fake_info_path, fake_info)
        else:
            fake_upper_snap_name = os.path.basename(fake_upper_snap_path)

            fake_backing_chain = [
                {'filename': fake_upper_snap_name,
                 'backing-filename': fake_snapshot_name},
                {'filename': fake_snapshot_name,
                 'backing-filename': volume_name},
                {'filename': volume_name,
                 'backing-filename': None}]

            fake_info[fake_upper_snap_id] = fake_upper_snap_name
            fake_info[self._fake_snapshot.id] = fake_snapshot_name
            fake_info['active'] = fake_upper_snap_name

            expected_info = copy.deepcopy(fake_info)
            del expected_info[snapshot.id]

            self._driver._read_info_file.return_value = fake_info
            self._driver._get_backing_chain_for_path = mock.Mock(
                return_value=fake_backing_chain)

            self._driver._delete_snapshot(snapshot)

            self._driver._img_commit.assert_called_once_with(
                snapshot_path)
            self._driver._rebase_img.assert_called_once_with(
                fake_upper_snap_path, volume_name,
                fake_base_img_info.file_format)
            self._driver._write_info_file.assert_called_once_with(
                mock.sentinel.fake_info_path, expected_info)

    @ddt.data({'encryption': True}, {'encryption': False})
    def test_delete_snapshot_when_active_file(self, encryption):
        self._test_delete_snapshot(encryption=encryption)

    @ddt.data({'encryption': True}, {'encryption': False})
    def test_delete_snapshot_in_use(self, encryption):
        self._test_delete_snapshot(volume_in_use=True,
                                   encryption=encryption)

    @ddt.data({'encryption': True}, {'encryption': False})
    def test_delete_snapshot_in_use_stale_snapshot(self,
                                                   encryption):
        self._test_delete_snapshot(volume_in_use=True,
                                   stale_snapshot=True,
                                   encryption=encryption)

    @ddt.data({'encryption': True}, {'encryption': False})
    def test_delete_snapshot_with_one_upper_file(self,
                                                 encryption):
        self._test_delete_snapshot(is_active_image=False,
                                   encryption=encryption)

    @ddt.data({'encryption': True}, {'encryption': False})
    def test_delete_stale_snapshot(self, encryption):
        if encryption:
            fake_snapshot_name = os.path.basename(self._fake_snap_c_path)
            volume_name = self.volume_c.name
            snapshot = self._fake_snap_c
            snapshot_path = self._fake_snap_c_path
        else:
            fake_snapshot_name = os.path.basename(self._fake_snapshot_path)
            volume_name = self._fake_volume.name
            snapshot = self._fake_snapshot
            snapshot_path = self._fake_snapshot_path

        fake_snap_info = {
            'active': volume_name,
            snapshot.id: fake_snapshot_name
        }
        expected_info = {'active': volume_name}

        self._driver._local_path_volume_info = mock.Mock(
            return_value=mock.sentinel.fake_info_path)
        self._driver._read_info_file = mock.Mock(
            return_value=fake_snap_info)
        self._driver._local_volume_dir = mock.Mock(
            return_value=self._FAKE_MNT_POINT)
        self._driver._write_info_file = mock.Mock()

        self._driver._delete_stale_snapshot(snapshot)

        self._driver._delete.assert_called_once_with(snapshot_path)
        self._driver._write_info_file.assert_called_once_with(
            mock.sentinel.fake_info_path, expected_info)

    @mock.patch.object(remotefs.RemoteFSDriver,
                       'secure_file_operations_enabled',
                       return_value=True)
    @mock.patch.object(os, 'stat')
    def test_do_create_snapshot(self, _mock_stat, _mock_sec_enabled):
        self._driver._local_volume_dir = mock.Mock(
            return_value=self._fake_volume_path)
        fake_backing_path = os.path.join(
            self._driver._local_volume_dir(),
            self._fake_volume.name)

        self._driver._execute = mock.Mock()
        self._driver._set_rw_permissions = mock.Mock()
        self._driver._qemu_img_info = mock.Mock(
            return_value=mock.Mock(file_format=mock.sentinel.backing_fmt))

        self._driver._do_create_snapshot(self._fake_snapshot,
                                         self._fake_volume.name,
                                         self._fake_snapshot_path)
        command1 = ['qemu-img', 'create', '-f', 'qcow2', '-o',
                    'backing_file=%s,backing_fmt=%s' %
                    (fake_backing_path,
                     mock.sentinel.backing_fmt),
                    self._fake_snapshot_path,
                    "%dG" % self._fake_volume.size]
        command2 = ['qemu-img', 'rebase', '-u',
                    '-b', self._fake_volume.name,
                    '-F', mock.sentinel.backing_fmt,
                    self._fake_snapshot_path]
        command3 = ['chown', '--reference=%s' % fake_backing_path,
                    self._fake_snapshot_path]
        calls = [mock.call(*command1, run_as_root=True),
                 mock.call(*command2, run_as_root=True),
                 mock.call(*command3, run_as_root=True)]
        self._driver._execute.assert_has_calls(calls)

    def _test_create_snapshot(self, display_name=None, volume_in_use=False,
                              encryption=False):
        fake_snapshot_info = {}
        if encryption:
            fake_snapshot_file_name = os.path.basename(self._fake_snap_c_path)
            volume_name = self.volume_c.name
            snapshot = self._fake_snap_c
            snapshot_path = self._fake_snap_c_path
        else:
            fake_snapshot_file_name = os.path.basename(
                self._fake_snapshot_path)
            volume_name = self._fake_volume.name
            snapshot = self._fake_snapshot
            snapshot_path = self._fake_snapshot_path

        snapshot.display_name = display_name
        self._driver._local_path_volume_info = mock.Mock(
            return_value=mock.sentinel.fake_info_path)
        self._driver._read_info_file = mock.Mock(
            return_value=fake_snapshot_info)
        self._driver._do_create_snapshot = mock.Mock()
        self._driver._create_snapshot_online = mock.Mock()
        self._driver._write_info_file = mock.Mock()
        self._driver.get_active_image_from_info = mock.Mock(
            return_value=volume_name)
        self._driver._get_new_snap_path = mock.Mock(
            return_value=snapshot_path)
        self._driver._validate_state = mock.Mock()

        expected_snapshot_info = {
            'active': fake_snapshot_file_name,
            snapshot.id: fake_snapshot_file_name
        }
        exp_acceptable_states = ['available', 'in-use', 'backing-up']
        if display_name and display_name.startswith('tmp-snap-'):
            exp_acceptable_states.append('downloading')
            self._fake_snapshot.volume.status = 'downloading'

        if volume_in_use:
            snapshot.volume.status = 'backing-up'
            snapshot.volume.attach_status = 'attached'
            expected_method_called = '_create_snapshot_online'
            conn_info = ('{"driver_volume_type": "nfs",'
                         '"export": "localhost:/srv/nfs1",'
                         '"name": "old_name"}')
            attachment = fake_volume.volume_attachment_ovo(
                self.context, connection_info=conn_info)
            snapshot.volume.volume_attachment.objects.append(attachment)
            mock_save = self.mock_object(attachment, 'save')

            # After the snapshot the connection info should change the name of
            # the file
            expected = copy.deepcopy(attachment.connection_info)
            expected['name'] = snapshot.volume.name + '.' + snapshot.id
        else:
            expected_method_called = '_do_create_snapshot'

        self._driver._create_snapshot(snapshot)

        self._driver._validate_state.assert_called_once_with(
            snapshot.volume.status,
            exp_acceptable_states)
        fake_method = getattr(self._driver, expected_method_called)
        fake_method.assert_called_with(
            snapshot, volume_name,
            snapshot_path)
        self._driver._write_info_file.assert_called_with(
            mock.sentinel.fake_info_path,
            expected_snapshot_info)

        if volume_in_use:
            mock_save.assert_called_once()
            changed_fields = attachment.cinder_obj_get_changes()
            self.assertEqual(expected, changed_fields['connection_info'])

    @ddt.data({'encryption': True}, {'encryption': False})
    def test_create_snapshot_volume_available(self, encryption):
        self._test_create_snapshot(encryption=encryption)

    @ddt.data({'encryption': True}, {'encryption': False})
    def test_create_snapshot_volume_in_use(self, encryption):
        self._test_create_snapshot(volume_in_use=True,
                                   encryption=encryption)

    def test_create_snapshot_invalid_volume(self):
        self._fake_snapshot.volume.status = 'error'
        self.assertRaises(exception.InvalidVolume,
                          self._driver._create_snapshot,
                          self._fake_snapshot)

    @ddt.data(None, 'test', 'tmp-snap-404f-404')
    def test_create_snapshot_names(self, display_name):
        self._test_create_snapshot(display_name=display_name)

    @mock.patch('cinder.db.snapshot_get')
    @mock.patch('time.sleep')
    def test_create_snapshot_online_with_concurrent_delete(
            self, mock_sleep, mock_snapshot_get):
        self._driver._nova = mock.Mock()

        # Test what happens when progress is so slow that someone
        # decides to delete the snapshot while the last known status is
        # "creating".
        mock_snapshot_get.side_effect = [
            {'status': 'creating', 'progress': '42%'},
            {'status': 'creating', 'progress': '45%'},
            {'status': 'deleting'},
        ]
        fake_snapshot = self._fake_snapshot
        fake_snapshot.context = self.context

        with mock.patch.object(self._driver, '_do_create_snapshot') as \
                mock_do_create_snapshot:
            self.assertRaises(exception.RemoteFSConcurrentRequest,
                              self._driver._create_snapshot_online,
                              fake_snapshot,
                              self._fake_volume.name,
                              self._fake_snapshot_path)

        mock_do_create_snapshot.assert_called_once_with(
            fake_snapshot, self._fake_volume.name,
            self._fake_snapshot_path)
        self.assertEqual([mock.call(1), mock.call(1)],
                         mock_sleep.call_args_list)
        self.assertEqual(3, mock_snapshot_get.call_count)
        mock_snapshot_get.assert_called_with(self._fake_snapshot._context,
                                             self._fake_snapshot.id)

    @mock.patch.object(utils, 'synchronized')
    def _locked_volume_operation_test_helper(self, mock_synchronized, func,
                                             expected_exception=False,
                                             *args, **kwargs):
        def mock_decorator(*args, **kwargs):
            def mock_inner(f):
                return f
            return mock_inner

        mock_synchronized.side_effect = mock_decorator
        expected_lock = '%s-%s' % (self._driver.driver_prefix,
                                   self._fake_volume.id)

        if expected_exception:
            self.assertRaises(expected_exception, func,
                              self._driver,
                              *args, **kwargs)
        else:
            ret_val = func(self._driver, *args, **kwargs)

            mock_synchronized.assert_called_with(expected_lock,
                                                 external=False)
            self.assertEqual(mock.sentinel.ret_val, ret_val)

    def test_locked_volume_id_operation(self):
        mock_volume = mock.Mock()
        mock_volume.id = self._fake_volume.id

        @remotefs.locked_volume_id_operation
        def synchronized_func(inst, volume):
            return mock.sentinel.ret_val

        self._locked_volume_operation_test_helper(func=synchronized_func,
                                                  volume=mock_volume)

    def test_locked_volume_id_snapshot_operation(self):
        mock_snapshot = mock.Mock()
        mock_snapshot.volume.id = self._fake_volume.id

        @remotefs.locked_volume_id_operation
        def synchronized_func(inst, snapshot):
            return mock.sentinel.ret_val

        self._locked_volume_operation_test_helper(func=synchronized_func,
                                                  snapshot=mock_snapshot)

    def test_locked_volume_id_operation_exception(self):
        @remotefs.locked_volume_id_operation
        def synchronized_func(inst):
            return mock.sentinel.ret_val

        self._locked_volume_operation_test_helper(
            func=synchronized_func,
            expected_exception=exception.VolumeBackendAPIException)

    @mock.patch.object(image_utils, 'qemu_img_info')
    @mock.patch('os.path.basename')
    def _test_qemu_img_info(self, mock_basename,
                            mock_qemu_img_info, backing_file, basedir,
                            template=None, valid_backing_file=True):
        fake_vol_name = 'fake_vol_name'
        mock_info = mock_qemu_img_info.return_value
        mock_info.image = mock.sentinel.image_path
        mock_info.backing_file = backing_file

        self._driver._VALID_IMAGE_EXTENSIONS = ['vhd', 'vhdx', 'raw', 'qcow2']

        mock_basename.side_effect = [mock.sentinel.image_basename,
                                     mock.sentinel.backing_file_basename]

        if valid_backing_file:
            img_info = self._driver._qemu_img_info_base(
                mock.sentinel.image_path, fake_vol_name, basedir,
                ext_bf_template=template)
            self.assertEqual(mock_info, img_info)
            self.assertEqual(mock.sentinel.image_basename,
                             mock_info.image)
            expected_basename_calls = [mock.call(mock.sentinel.image_path)]
            if backing_file:
                self.assertEqual(mock.sentinel.backing_file_basename,
                                 mock_info.backing_file)
                expected_basename_calls.append(mock.call(backing_file))
            mock_basename.assert_has_calls(expected_basename_calls)
        else:
            self.assertRaises(exception.RemoteFSInvalidBackingFile,
                              self._driver._qemu_img_info_base,
                              mock.sentinel.image_path,
                              fake_vol_name, basedir)

        mock_qemu_img_info.assert_called_with(mock.sentinel.image_path,
                                              force_share=False,
                                              run_as_root=True)

    @ddt.data([None, '/fake_basedir'],
              ['/fake_basedir/cb2016/fake_vol_name', '/fake_basedir'],
              ['/fake_basedir/cb2016/fake_vol_name.VHD', '/fake_basedir'],
              ['/fake_basedir/cb2016/fake_vol_name.404f-404',
               '/fake_basedir'],
              ['/fake_basedir/cb2016/fake_vol_name.tmp-snap-404f-404',
               '/fake_basedir'])
    @ddt.unpack
    def test_qemu_img_info_valid_backing_file(self, backing_file, basedir):
        self._test_qemu_img_info(backing_file=backing_file,
                                 basedir=basedir)

    @ddt.data(['/other_random_path', '/fake_basedir'],
              ['/other_basedir/cb2016/fake_vol_name', '/fake_basedir'],
              ['/fake_basedir/invalid_hash/fake_vol_name', '/fake_basedir'],
              ['/fake_basedir/cb2016/invalid_vol_name', '/fake_basedir'],
              ['/fake_basedir/cb2016/fake_vol_name.info', '/fake_basedir'],
              ['/fake_basedir/cb2016/fake_vol_name-random-suffix',
               '/fake_basedir'],
              ['/fake_basedir/cb2016/fake_vol_name.invalidext',
               '/fake_basedir'])
    @ddt.unpack
    def test_qemu_img_info_invalid_backing_file(self, backing_file, basedir):
        self._test_qemu_img_info(backing_file=backing_file,
                                 basedir=basedir,
                                 valid_backing_file=False)

    @ddt.data([None, '/fake_basedir'],
              ['/fake_basedir/cb2016/fake_vol_name', '/fake_basedir'],
              ['/fake_basedir/cb2016/fake_vol_name.VHD', '/fake_basedir'],
              ['/fake_basedir/cb2016/fake_vol_name.404f-404',
               '/fake_basedir'],
              ['/fake_basedir/cb2016/fake_vol_name.tmp-snap-404f-404',
               '/fake_basedir'],
              ['/fake_basedir/cb2016/other_dir/404f-404',
               '/fake_basedir'],
              ['/fake_basedir/cb2016/other_dir/tmp-snap-404f-404',
               '/fake_basedir'],
              ['/fake_basedir/cb2016/other_dir/404f-404.mod1-404f-404',
               '/fake_basedir'],
              ['/fake_basedir/cb2016/other_dir/404f-404.mod2-404f-404',
               '/fake_basedir'])
    @ddt.unpack
    def test_qemu_img_info_extended_backing_file(self, backing_file, basedir):
        """Tests using a special backing file template

        The special backing file template used in here allows backing files
        in a subdirectory and with special extended names (.mod1-[], .mod2-[],
        ...).
        """
        ext_template = ("(#basedir/[0-9a-f]+/)?(#volname(.(tmp-snap-)"
                        "?[0-9a-f-]+)?#valid_ext|other_dir/(tmp-snap-)?"
                        "[0-9a-f-]+(.(mod1-|mod2-)[0-9a-f-]+)?)$")
        self._test_qemu_img_info(backing_file=backing_file,
                                 basedir=basedir,
                                 template=remotefs.BackingFileTemplate(
                                     ext_template),
                                 valid_backing_file=True)

    @ddt.data(['/other_random_path', '/fake_basedir'],
              ['/other_basedir/cb2016/fake_vol_name', '/fake_basedir'],
              ['/fake_basedir/invalid_hash/fake_vol_name', '/fake_basedir'],
              ['/fake_basedir/cb2016/invalid_vol_name', '/fake_basedir'],
              ['/fake_basedir/cb2016/fake_vol_name.info', '/fake_basedir'],
              ['/fake_basedir/cb2016/fake_vol_name-random-suffix',
               '/fake_basedir'],
              ['/fake_basedir/cb2016/fake_vol_name.invalidext',
               '/fake_basedir'],
              ['/fake_basedir/cb2016/invalid_dir/404f-404',
               '/fake_basedir'],
              ['/fake_basedir/cb2016/other_dir/invalid-prefix-404f-404',
               '/fake_basedir'],
              ['/fake_basedir/cb2016/other_dir/404f-404.mod3-404f-404',
               '/fake_basedir'],
              ['/fake_basedir/cb2016/other_dir/404f-404.mod2-404f-404.invalid',
               '/fake_basedir'])
    @ddt.unpack
    def test_qemu_img_info_extended_backing_file_invalid(self, backing_file,
                                                         basedir):
        """Tests using a special backing file template with invalid files

        The special backing file template used in here allows backing files
        in a subdirectory and with special extended names (.mod1-[], .mod2-[],
        ...).
        """
        ext_template = ("(#basedir/[0-9a-f]+/)?(#volname(.(tmp-snap-)"
                        "?[0-9a-f-]+)?#valid_ext|other_dir/(tmp-snap-)?"
                        "[0-9a-f-]+(.(mod1-|mod2-)[0-9a-f-]+)?)$")
        self._test_qemu_img_info(backing_file=backing_file,
                                 basedir=basedir,
                                 template=remotefs.BackingFileTemplate(
                                     ext_template),
                                 valid_backing_file=False)

    @mock.patch.object(remotefs.RemoteFSSnapDriver, '_local_volume_dir')
    @mock.patch.object(remotefs.RemoteFSSnapDriver,
                       'get_active_image_from_info')
    def test_local_path_active_image(self, mock_get_active_img,
                                     mock_local_vol_dir):
        fake_vol_dir = 'fake_vol_dir'
        fake_active_img = 'fake_active_img_fname'

        mock_get_active_img.return_value = fake_active_img
        mock_local_vol_dir.return_value = fake_vol_dir

        active_img_path = self._driver._local_path_active_image(
            mock.sentinel.volume)
        exp_act_img_path = os.path.join(fake_vol_dir, fake_active_img)

        self.assertEqual(exp_act_img_path, active_img_path)
        mock_get_active_img.assert_called_once_with(mock.sentinel.volume)
        mock_local_vol_dir.assert_called_once_with(mock.sentinel.volume)

    @ddt.data({},
              {'provider_location': None},
              {'active_fpath': 'last_snap_img',
               'expect_snaps': True})
    @ddt.unpack
    @mock.patch.object(remotefs.RemoteFSSnapDriver,
                       '_local_path_active_image')
    @mock.patch.object(remotefs.RemoteFSSnapDriver,
                       'local_path')
    def test_snapshots_exist(self, mock_local_path,
                             mock_local_path_active_img,
                             provider_location='fake_share',
                             active_fpath='base_img_path',
                             base_vol_path='base_img_path',
                             expect_snaps=False):
        self._fake_volume.provider_location = provider_location

        mock_local_path.return_value = base_vol_path
        mock_local_path_active_img.return_value = active_fpath

        snaps_exist = self._driver._snapshots_exist(self._fake_volume)

        self.assertEqual(expect_snaps, snaps_exist)

        if provider_location:
            mock_local_path.assert_called_once_with(self._fake_volume)
            mock_local_path_active_img.assert_called_once_with(
                self._fake_volume)
        else:
            self.assertFalse(mock_local_path.called)

    @ddt.data({},
              {'snapshots_exist': True},
              {'force_temp_snap': True})
    @ddt.unpack
    @mock.patch.object(sys.modules['cinder.objects'], "Snapshot")
    @mock.patch.object(remotefs.RemoteFSSnapDriver, 'local_path')
    @mock.patch.object(remotefs.RemoteFSSnapDriver, '_snapshots_exist')
    @mock.patch.object(remotefs.RemoteFSSnapDriver, '_copy_volume_image')
    @mock.patch.object(remotefs.RemoteFSSnapDriver, '_extend_volume')
    @mock.patch.object(remotefs.RemoteFSSnapDriver, '_validate_state')
    @mock.patch.object(remotefs.RemoteFSSnapDriver, '_create_snapshot')
    @mock.patch.object(remotefs.RemoteFSSnapDriver, '_delete_snapshot')
    @mock.patch.object(remotefs.RemoteFSSnapDriver,
                       '_copy_volume_from_snapshot')
    def test_create_cloned_volume(self, mock_copy_volume_from_snapshot,
                                  mock_delete_snapshot,
                                  mock_create_snapshot,
                                  mock_validate_state,
                                  mock_extend_volume,
                                  mock_copy_volume_image,
                                  mock_snapshots_exist,
                                  mock_local_path,
                                  mock_obj_snap,
                                  snapshots_exist=False,
                                  force_temp_snap=False):
        drv = self._driver

        # prepare test
        volume = fake_volume.fake_volume_obj(self.context)
        src_vref_id = '375e32b2-804a-49f2-b282-85d1d5a5b9e1'
        src_vref = fake_volume.fake_volume_obj(
            self.context,
            id=src_vref_id,
            name='volume-%s' % src_vref_id,
            obj_context=self.context)
        src_vref.context = self.context

        mock_snapshots_exist.return_value = snapshots_exist
        drv._always_use_temp_snap_when_cloning = force_temp_snap

        vol_attrs = ['provider_location', 'size', 'id', 'name', 'status',
                     'volume_type', 'metadata', 'obj_context']
        Volume = collections.namedtuple('Volume', vol_attrs)

        volume_ref = Volume(id=volume.id,
                            metadata=volume.metadata,
                            name=volume.name,
                            provider_location=volume.provider_location,
                            status=volume.status,
                            size=volume.size,
                            volume_type=volume.volume_type,
                            obj_context=self.context,)

        snap_args_creation = {
            'volume_id': src_vref.id,
            'user_id': None,
            'project_id': None,
            'status': fields.SnapshotStatus.CREATING,
            'progress': '0%',
            'volume_size': src_vref.size,
            'display_name': 'tmp-snap-%s' % volume.id,
            'display_description': None,
            'volume_type_id': src_vref.volume_type_id,
            'encryption_key_id': None,
        }
        snap_args_deletion = snap_args_creation.copy()
        snap_args_deletion["status"] = fields.SnapshotStatus.DELETED
        snap_args_deletion["deleted"] = True

        mock_obj_snap.return_value = mock.Mock()
        mock_obj_snap.return_value.create = mock.Mock()
        # end of prepare test

        # run test
        drv.create_cloned_volume(volume, src_vref)

        # evaluate test
        exp_acceptable_states = ['available', 'backing-up', 'downloading']
        mock_validate_state.assert_called_once_with(
            src_vref.status,
            exp_acceptable_states,
            obj_description='source volume')

        if snapshots_exist or force_temp_snap:
            mock_obj_snap.return_value.create.assert_called_once_with()
            mock_obj_snap.assert_called_once_with(
                context=self.context, **snap_args_creation)
            mock_create_snapshot.assert_called_once_with(
                mock_obj_snap.return_value)
            mock_copy_volume_from_snapshot.assert_called_once_with(
                mock_obj_snap.return_value, volume_ref, volume['size'],
                src_encryption_key_id=None, new_encryption_key_id=None)
            mock_delete_snapshot.called_once_with(snap_args_deletion)
        else:
            self.assertFalse(mock_create_snapshot.called)

            mock_snapshots_exist.assert_called_once_with(src_vref)

            mock_copy_volume_image.assert_called_once_with(
                mock_local_path.return_value,
                mock_local_path.return_value)
            mock_local_path.assert_has_calls(
                [mock.call(src_vref), mock.call(volume_ref)])
            mock_extend_volume.assert_called_once_with(volume_ref, volume.size)

    @ddt.data(None, 'raw', 'qcow2')
    @mock.patch('cinder.objects.volume.Volume.save')
    @mock.patch.object(sys.modules['cinder.objects'], "Snapshot")
    @mock.patch.object(remotefs.RemoteFSSnapDriver, 'local_path')
    @mock.patch.object(remotefs.RemoteFSSnapDriver, '_snapshots_exist')
    @mock.patch.object(remotefs.RemoteFSSnapDriver, '_copy_volume_image')
    @mock.patch.object(remotefs.RemoteFSSnapDriver, '_extend_volume')
    @mock.patch.object(remotefs.RemoteFSSnapDriver, '_validate_state')
    @mock.patch.object(remotefs.RemoteFSSnapDriver, '_create_snapshot')
    @mock.patch.object(remotefs.RemoteFSSnapDriver, '_delete_snapshot')
    @mock.patch.object(remotefs.RemoteFSSnapDriver,
                       '_copy_volume_from_snapshot')
    def test_create_cloned_volume_with_format(
            self, file_format, mock_copy_volume_from_snapshot,
            mock_delete_snapshot, mock_create_snapshot,
            mock_validate_state, mock_extend_volume,
            mock_copy_volume_image, mock_snapshots_exist,
            mock_local_path, mock_obj_snap, mock_save):
        drv = self._driver

        # prepare test
        volume = fake_volume.fake_volume_obj(self.context)
        src_vref_id = '375e32b2-804a-49f2-b282-85d1d5a5b9e1'
        src_vref = fake_volume.fake_volume_obj(
            self.context,
            id=src_vref_id,
            name='volume-%s' % src_vref_id,
            obj_context=self.context)
        src_vref.context = self.context
        if file_format:
            src_vref.admin_metadata = {'format': file_format}

        mock_snapshots_exist.return_value = False
        drv._always_use_temp_snap_when_cloning = False

        vol_attrs = ['provider_location', 'size', 'id', 'name', 'status',
                     'volume_type', 'metadata', 'obj_context']
        Volume = collections.namedtuple('Volume', vol_attrs)

        volume_ref = Volume(id=volume.id,
                            metadata=volume.metadata,
                            name=volume.name,
                            provider_location=volume.provider_location,
                            status=volume.status,
                            size=volume.size,
                            volume_type=volume.volume_type,
                            obj_context=self.context,)

        snap_args_creation = {
            'volume_id': src_vref.id,
            'user_id': None,
            'project_id': None,
            'status': fields.SnapshotStatus.CREATING,
            'progress': '0%',
            'volume_size': src_vref.size,
            'display_name': 'tmp-snap-%s' % volume.id,
            'display_description': None,
            'volume_type_id': src_vref.volume_type_id,
            'encryption_key_id': None,
        }
        snap_args_deletion = snap_args_creation.copy()
        snap_args_deletion["status"] = fields.SnapshotStatus.DELETED
        snap_args_deletion["deleted"] = True

        mock_obj_snap.return_value = mock.Mock()
        mock_obj_snap.return_value.create = mock.Mock()
        # end of prepare test

        # run test
        drv.create_cloned_volume(volume, src_vref)

        # evaluate test
        exp_acceptable_states = ['available', 'backing-up', 'downloading']
        mock_validate_state.assert_called_once_with(
            src_vref.status,
            exp_acceptable_states,
            obj_description='source volume')

        self.assertFalse(mock_create_snapshot.called)

        mock_snapshots_exist.assert_called_once_with(src_vref)

        mock_copy_volume_image.assert_called_once_with(
            mock_local_path.return_value,
            mock_local_path.return_value)
        mock_local_path.assert_has_calls(
            [mock.call(src_vref), mock.call(volume_ref)])
        mock_extend_volume.assert_called_once_with(volume_ref, volume.size)
        if file_format:
            self.assertEqual(file_format,
                             volume.admin_metadata['format'])

    @mock.patch('tempfile.NamedTemporaryFile')
    @mock.patch('cinder.volume.volume_utils.check_encryption_provider',
                return_value={'encryption_key_id': fake.ENCRYPTION_KEY_ID})
    def test_create_encrypted_volume(self,
                                     mock_check_enc_prov,
                                     mock_temp_file):
        class DictObj(object):
            # convert a dict to object w/ attributes
            def __init__(self, d):
                self.__dict__ = d

        drv = self._driver

        mock_temp_file.return_value.__enter__.side_effect = [
            DictObj({'name': '/imgfile'}),
            DictObj({'name': '/passfile'})]

        key_mgr = fake_keymgr.fake_api()

        self.mock_object(castellan.key_manager, 'API', return_value=key_mgr)
        key_id = key_mgr.store(self.context, KeyObject())
        self.volume_c.encryption_key_id = key_id

        enc_info = {'encryption_key_id': key_id,
                    'cipher': 'aes-xts-essiv',
                    'key_size': 256}

        remotefs_path = 'cinder.volume.drivers.remotefs.open'
        with mock.patch('cinder.volume.volume_utils.check_encryption_provider',
                        return_value=enc_info), \
                mock.patch(remotefs_path) as mock_open, \
                mock.patch.object(drv, '_execute') as mock_exec:

            drv._create_encrypted_volume_file("/passfile",
                                              self.volume_c.size,
                                              enc_info,
                                              self.context)

            mock_open.assert_called_with('/imgfile', 'w')
            mock_exec.assert_called()

    @mock.patch('shutil.copyfile')
    @mock.patch.object(remotefs.RemoteFSSnapDriver, '_set_rw_permissions')
    def test_copy_volume_image(self, mock_set_perm, mock_copyfile):
        self._driver._copy_volume_image(mock.sentinel.src, mock.sentinel.dest)

        mock_copyfile.assert_called_once_with(mock.sentinel.src,
                                              mock.sentinel.dest)
        mock_set_perm.assert_called_once_with(mock.sentinel.dest)

    def test_create_regular_file(self):
        self._driver._create_regular_file('/path', 1)
        self._driver._execute.assert_called_once_with('dd', 'if=/dev/zero',
                                                      'of=/path', 'bs=1M',
                                                      'count=1024',
                                                      run_as_root=True)

    @mock.patch.object(remotefs.RemoteFSSnapDriver, '_local_path_volume_info')
    @mock.patch.object(remotefs.RemoteFSSnapDriver, '_read_info_file')
    @mock.patch.object(remotefs.RemoteFSSnapDriver, '_local_volume_dir')
    @mock.patch.object(remotefs.RemoteFSSnapDriver, '_qemu_img_info')
    def test_get_snapshot_backing_file(
            self, mock_qemu_img_info, mock_local_vol_dir,
            mock_read_info_file, mock_local_path_vol_info):

        fake_snapshot_file_name = os.path.basename(self._fake_snapshot_path)
        fake_snapshot_info = {self._fake_snapshot.id: fake_snapshot_file_name}

        fake_snap_img_info = mock.Mock()
        fake_snap_img_info.backing_file = self._fake_volume.name

        mock_read_info_file.return_value = fake_snapshot_info
        mock_qemu_img_info.return_value = fake_snap_img_info
        mock_local_vol_dir.return_value = self._FAKE_MNT_POINT

        snap_backing_file = self._driver._get_snapshot_backing_file(
            self._fake_snapshot)
        self.assertEqual(os.path.basename(self._fake_volume_path),
                         snap_backing_file)

        mock_local_path_vol_info.assert_called_once_with(self._fake_volume)
        mock_read_info_file.assert_called_once_with(
            mock_local_path_vol_info.return_value)
        mock_local_vol_dir.assert_called_once_with(self._fake_volume)
        mock_qemu_img_info.assert_called_once_with(self._fake_snapshot_path)

    @ddt.data({},
              {'info_file_exists': True},
              {'os_name': 'nt'})
    @ddt.unpack
    @mock.patch('json.dump')
    @mock.patch('cinder.volume.drivers.remotefs.open')
    @mock.patch('os.path.exists')
    def test_write_info_file(self,
                             mock_os_path_exists,
                             mock_open,
                             mock_json_dump,
                             info_file_exists=False,
                             os_name='posix'):

        mock_os_path_exists.return_value = info_file_exists
        fake_info_path = '/path/to/info'
        fake_snapshot_info = {'active': self._fake_snapshot_path}
        self._driver._execute = mock.Mock()
        self._driver._set_rw_permissions = mock.Mock()

        self._driver._write_info_file(fake_info_path, fake_snapshot_info)

        mock_open.assert_called_once_with(fake_info_path, 'w')
        mock_json_dump.assert_called_once_with(
            fake_snapshot_info, mock.ANY, indent=1, sort_keys=True)

        if info_file_exists or os.name == 'nt':
            self._driver._execute.assert_not_called()
            self._driver._set_rw_permissions.assert_not_called()
        else:
            self._driver._execute.assert_called_once_with(
                'truncate', "-s0", fake_info_path,
                run_as_root=self._driver._execute_as_root)
            self._driver._set_rw_permissions.assert_called_once_with(
                fake_info_path)

        fake_snapshot_info.pop('active')
        self.assertRaises(exception.RemoteFSException,
                          self._driver._write_info_file,
                          fake_info_path,
                          fake_snapshot_info)


class RemoteFSPoolMixinTestCase(test.TestCase):
    def setUp(self):
        super(RemoteFSPoolMixinTestCase, self).setUp()
        # We'll instantiate this directly for now.
        self._driver = remotefs.RemoteFSPoolMixin()

        self.context = context.get_admin_context()

    @mock.patch.object(remotefs.RemoteFSPoolMixin,
                       '_get_pool_name_from_volume')
    @mock.patch.object(remotefs.RemoteFSPoolMixin,
                       '_get_share_from_pool_name')
    def test_find_share(self, mock_get_share_from_pool,
                        mock_get_pool_from_volume):
        share = self._driver._find_share(mock.sentinel.volume)

        self.assertEqual(mock_get_share_from_pool.return_value, share)
        mock_get_pool_from_volume.assert_called_once_with(
            mock.sentinel.volume)
        mock_get_share_from_pool.assert_called_once_with(
            mock_get_pool_from_volume.return_value)

    def test_get_pool_name_from_volume(self):
        fake_pool = 'fake_pool'
        fake_host = 'fake_host@fake_backend#%s' % fake_pool
        fake_vol = fake_volume.fake_volume_obj(
            self.context, provider_location='fake_share',
            host=fake_host)

        pool_name = self._driver._get_pool_name_from_volume(fake_vol)
        self.assertEqual(fake_pool, pool_name)

    def test_update_volume_stats(self):
        share_total_gb = 3
        share_free_gb = 2
        share_used_gb = 4  # provisioned space

        self._driver._mounted_shares = [mock.sentinel.share]

        self._driver.configuration = mock.Mock()
        self._driver.configuration.safe_get.return_value = (
            mock.sentinel.backend_name)
        self._driver.vendor_name = mock.sentinel.vendor_name
        self._driver.driver_volume_type = mock.sentinel.driver_volume_type
        self._driver._thin_provisioning_support = (
            mock.sentinel.thin_prov_support)
        self._driver._thick_provisioning_support = (
            mock.sentinel.thick_prov_support)

        self._driver.get_version = mock.Mock(
            return_value=mock.sentinel.driver_version)
        self._driver._ensure_shares_mounted = mock.Mock()
        self._driver._get_capacity_info = mock.Mock(
            return_value=(share_total_gb << 30,
                          share_free_gb << 30,
                          share_used_gb << 30))
        self._driver._get_pool_name_from_share = mock.Mock(
            return_value=mock.sentinel.pool_name)

        expected_pool = {
            'pool_name': mock.sentinel.pool_name,
            'total_capacity_gb': float(share_total_gb),
            'free_capacity_gb': float(share_free_gb),
            'provisioned_capacity_gb': float(share_used_gb),
            'reserved_percentage': (
                self._driver.configuration.reserved_percentage),
            'max_over_subscription_ratio': (
                self._driver.configuration.max_over_subscription_ratio),
            'thin_provisioning_support': (
                mock.sentinel.thin_prov_support),
            'thick_provisioning_support': (
                mock.sentinel.thick_prov_support),
            'QoS_support': False,
        }

        expected_stats = {
            'volume_backend_name': mock.sentinel.backend_name,
            'vendor_name': mock.sentinel.vendor_name,
            'driver_version': mock.sentinel.driver_version,
            'storage_protocol': mock.sentinel.driver_volume_type,
            'total_capacity_gb': 0,
            'free_capacity_gb': 0,
            'pools': [expected_pool],
        }

        self._driver._update_volume_stats()

        self.assertDictEqual(expected_stats, self._driver._stats)

        self._driver._get_capacity_info.assert_called_once_with(
            mock.sentinel.share)
        self._driver.configuration.safe_get.assert_called_once_with(
            'volume_backend_name')


@ddt.ddt
class RevertToSnapshotMixinTestCase(test.TestCase):

    _FAKE_MNT_POINT = '/mnt/fake_hash'

    def setUp(self):
        super(RevertToSnapshotMixinTestCase, self).setUp()
        self._driver = remotefs.RevertToSnapshotMixin()
        self._driver._remotefsclient = mock.Mock()
        self._driver._execute = mock.Mock()
        self._driver._delete = mock.Mock()

        self.context = context.get_admin_context()

        self._fake_volume = fake_volume.fake_volume_obj(
            self.context, provider_location='fake_share')
        self._fake_volume_path = os.path.join(self._FAKE_MNT_POINT,
                                              self._fake_volume.name)
        self._fake_snapshot = fake_snapshot.fake_snapshot_obj(self.context)
        self._fake_snapshot_path = (self._fake_volume_path + '.' +
                                    self._fake_snapshot.id)
        self._fake_snapshot_name = os.path.basename(
            self._fake_snapshot_path)
        self._fake_snapshot.volume = self._fake_volume

    @ddt.data(True, False)
    @mock.patch.object(remotefs.RevertToSnapshotMixin, '_validate_state',
                       create=True)
    @mock.patch.object(remotefs.RevertToSnapshotMixin, '_read_info_file',
                       create=True)
    @mock.patch.object(remotefs.RevertToSnapshotMixin,
                       '_local_path_volume_info', create=True)
    @mock.patch.object(remotefs.RevertToSnapshotMixin, '_qemu_img_info',
                       create=True)
    @mock.patch.object(remotefs.RevertToSnapshotMixin, '_do_create_snapshot',
                       create=True)
    @mock.patch.object(remotefs.RevertToSnapshotMixin, '_local_volume_dir',
                       create=True)
    def test_revert_to_snapshot(self,
                                is_latest_snapshot,
                                mock_local_vol_dir,
                                mock_do_create_snapshot,
                                mock_qemu_img_info,
                                mock_local_path_vol_info,
                                mock_read_info_file,
                                mock_validate_state):

        active_file = (self._fake_snapshot_name if is_latest_snapshot
                       else 'fake_latest_snap')
        fake_snapshot_info = {
            'active': active_file,
            self._fake_snapshot.id: self._fake_snapshot_name
        }

        mock_read_info_file.return_value = fake_snapshot_info

        fake_snap_img_info = mock.Mock()
        fake_snap_img_info.backing_file = self._fake_volume.name

        mock_qemu_img_info.return_value = fake_snap_img_info
        mock_local_vol_dir.return_value = self._FAKE_MNT_POINT

        if is_latest_snapshot:
            self._driver._revert_to_snapshot(self.context, self._fake_volume,
                                             self._fake_snapshot)
            self._driver._delete.assert_called_once_with(
                self._fake_snapshot_path)
            mock_do_create_snapshot.assert_called_once_with(
                self._fake_snapshot,
                fake_snap_img_info.backing_file,
                self._fake_snapshot_path)
            mock_qemu_img_info.assert_called_once_with(
                self._fake_snapshot_path,
                self._fake_volume.name)
        elif not is_latest_snapshot:
            self.assertRaises(exception.InvalidSnapshot,
                              self._driver._revert_to_snapshot,
                              self.context, self._fake_volume,
                              self._fake_snapshot)
            self._driver._delete.assert_not_called()

        exp_acceptable_states = ['available', 'reverting']
        mock_validate_state.assert_called_once_with(
            self._fake_snapshot.volume.status,
            exp_acceptable_states)
        mock_local_path_vol_info.assert_called_once_with(
            self._fake_snapshot.volume)
        mock_read_info_file.assert_called_once_with(
            mock_local_path_vol_info.return_value)


@ddt.ddt
class RemoteFSManageableVolumesTestCase(test.TestCase):
    def setUp(self):
        super(RemoteFSManageableVolumesTestCase, self).setUp()
        # We'll instantiate this directly for now.
        self._driver = remotefs.RemoteFSManageableVolumesMixin()

    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       '_get_mount_point_for_share', create=True)
    @mock.patch.object(os.path, 'isfile')
    def test_get_manageable_vol_location_invalid(self, mock_is_file,
                                                 mock_get_mount_point):
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self._driver._get_manageable_vol_location,
                          {})

        self._driver._mounted_shares = []
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self._driver._get_manageable_vol_location,
                          {'source-name': '//hots/share/img'})

        self._driver._mounted_shares = ['//host/share']
        mock_get_mount_point.return_value = '/fake_mountpoint'
        mock_is_file.return_value = False

        self.assertRaises(exception.ManageExistingInvalidReference,
                          self._driver._get_manageable_vol_location,
                          {'source-name': '//host/share/subdir/img'})
        mock_is_file.assert_any_call(
            os.path.normpath('/fake_mountpoint/subdir/img'))

    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       '_get_mount_point_for_share', create=True)
    @mock.patch.object(os.path, 'isfile')
    def test_get_manageable_vol_location(self, mock_is_file,
                                         mock_get_mount_point):
        self._driver._mounted_shares = [
            '//host/share2/subdir',
            '//host/share/subdir',
            'host:/dir/subdir'
        ]

        mock_get_mount_point.return_value = '/fake_mountpoint'
        mock_is_file.return_value = True

        location_info = self._driver._get_manageable_vol_location(
            {'source-name': 'host:/dir/subdir/import/img'})

        exp_location_info = {
            'share': 'host:/dir/subdir',
            'mountpoint': mock_get_mount_point.return_value,
            'vol_local_path': '/fake_mountpoint/import/img',
            'vol_remote_path': 'host:/dir/subdir/import/img'
        }
        self.assertEqual(exp_location_info, location_info)

    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       '_get_mount_point_for_share', create=True)
    @mock.patch.object(os.path, 'isfile')
    @mock.patch.object(os.path, 'normpath', lambda x: x.replace('/', '\\'))
    @mock.patch.object(os.path, 'normcase', lambda x: x.lower())
    @mock.patch.object(os.path, 'join', lambda *args: '\\'.join(args))
    @mock.patch.object(os.path, 'sep', '\\')
    def test_get_manageable_vol_location_win32(self, mock_is_file,
                                               mock_get_mount_point):
        self._driver._mounted_shares = [
            '//host/share2/subdir',
            '//host/share/subdir',
            'host:/dir/subdir'
        ]

        mock_get_mount_point.return_value = r'c:\fake_mountpoint'
        mock_is_file.return_value = True

        location_info = self._driver._get_manageable_vol_location(
            {'source-name': '//Host/share/Subdir/import/img'})

        exp_location_info = {
            'share': '//host/share/subdir',
            'mountpoint': mock_get_mount_point.return_value,
            'vol_local_path': r'c:\fake_mountpoint\import\img',
            'vol_remote_path': r'\\host\share\subdir\import\img'
        }
        self.assertEqual(exp_location_info, location_info)

    def test_get_managed_vol_exp_path(self):
        fake_vol = fake_volume.fake_volume_obj(mock.sentinel.context)
        vol_location = dict(mountpoint='fake-mountpoint')

        exp_path = os.path.join(vol_location['mountpoint'],
                                fake_vol.name)
        ret_val = self._driver._get_managed_vol_expected_path(
            fake_vol, vol_location)
        self.assertEqual(exp_path, ret_val)

    @ddt.data(
        {'already_managed': True},
        {'qemu_side_eff': exception.RemoteFSInvalidBackingFile},
        {'qemu_side_eff': Exception},
        {'qemu_side_eff': [mock.Mock(backing_file=None,
                                     file_format='fakefmt')]},
        {'qemu_side_eff': [mock.Mock(backing_file='backing_file',
                                     file_format='raw')]}
    )
    @ddt.unpack
    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       '_qemu_img_info', create=True)
    def test_check_unmanageable_volume(self, mock_qemu_info,
                                       qemu_side_eff=None,
                                       already_managed=False):
        mock_qemu_info.side_effect = qemu_side_eff

        manageable = self._driver._is_volume_manageable(
            mock.sentinel.volume_path,
            already_managed=already_managed)[0]
        self.assertFalse(manageable)

    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       '_qemu_img_info', create=True)
    def test_check_manageable_volume(self, mock_qemu_info,
                                     qemu_side_eff=None,
                                     already_managed=False):
        mock_qemu_info.return_value = mock.Mock(
            backing_file=None,
            file_format='raw')

        manageable = self._driver._is_volume_manageable(
            mock.sentinel.volume_path)[0]
        self.assertTrue(manageable)

    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       '_get_manageable_vol_location')
    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       '_is_volume_manageable')
    def test_manage_existing_unmanageable(self, mock_check_manageable,
                                          mock_get_location):
        fake_vol = fake_volume.fake_volume_obj(mock.sentinel.context)

        mock_get_location.return_value = dict(
            vol_local_path=mock.sentinel.local_path)
        mock_check_manageable.return_value = False, mock.sentinel.resason

        self.assertRaises(exception.ManageExistingInvalidReference,
                          self._driver.manage_existing,
                          fake_vol,
                          mock.sentinel.existing_ref)
        mock_get_location.assert_called_once_with(mock.sentinel.existing_ref)
        mock_check_manageable.assert_called_once_with(
            mock.sentinel.local_path)

    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       '_get_manageable_vol_location')
    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       '_is_volume_manageable')
    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       '_set_rw_permissions', create=True)
    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       '_get_managed_vol_expected_path')
    @mock.patch.object(os, 'rename')
    def test_manage_existing_manageable(self, mock_rename,
                                        mock_get_exp_path,
                                        mock_set_perm,
                                        mock_check_manageable,
                                        mock_get_location):
        fake_vol = fake_volume.fake_volume_obj(mock.sentinel.context)

        mock_get_location.return_value = dict(
            vol_local_path=mock.sentinel.local_path,
            share=mock.sentinel.share)
        mock_check_manageable.return_value = True, None

        exp_ret_val = {'provider_location': mock.sentinel.share}
        ret_val = self._driver.manage_existing(fake_vol,
                                               mock.sentinel.existing_ref)
        self.assertEqual(exp_ret_val, ret_val)

        mock_get_exp_path.assert_called_once_with(
            fake_vol, mock_get_location.return_value)
        mock_set_perm.assert_called_once_with(mock.sentinel.local_path)
        mock_rename.assert_called_once_with(mock.sentinel.local_path,
                                            mock_get_exp_path.return_value)

    @mock.patch.object(image_utils, 'qemu_img_info')
    def _get_rounded_manageable_image_size(self, mock_qemu_info):
        mock_qemu_info.return_value.virtual_size = 1 << 30 + 1
        exp_rounded_size_gb = 2

        size = self._driver._get_rounded_manageable_image_size(
            mock.sentinel.image_path)
        self.assertEqual(exp_rounded_size_gb, size)

        mock_qemu_info.assert_called_once_with(mock.sentinel.image_path)

    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       '_get_manageable_vol_location')
    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       '_get_rounded_manageable_image_size')
    def test_manage_existing_get_size(self, mock_get_size,
                                      mock_get_location):
        mock_get_location.return_value = dict(
            vol_local_path=mock.sentinel.image_path)

        size = self._driver.manage_existing_get_size(
            mock.sentinel.volume,
            mock.sentinel.existing_ref)
        self.assertEqual(mock_get_size.return_value, size)

        mock_get_location.assert_called_once_with(mock.sentinel.existing_ref)
        mock_get_size.assert_called_once_with(mock.sentinel.image_path)

    @ddt.data(
        {},
        {'managed_volume': mock.Mock(size=mock.sentinel.sz),
         'exp_size': mock.sentinel.sz,
         'manageable_check_ret_val': False,
         'exp_manageable': False},
        {'exp_size': None,
         'get_size_side_effect': Exception,
         'exp_manageable': False})
    @ddt.unpack
    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       '_is_volume_manageable')
    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       '_get_rounded_manageable_image_size')
    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       '_get_mount_point_for_share', create=True)
    def test_get_manageable_volume(
            self, mock_get_mount_point,
            mock_get_size, mock_check_manageable,
            managed_volume=None,
            get_size_side_effect=(mock.sentinel.size_gb, ),
            manageable_check_ret_val=True,
            exp_size=mock.sentinel.size_gb,
            exp_manageable=True):
        share = '//host/share'
        mountpoint = '/fake-mountpoint'
        volume_path = '/fake-mountpoint/subdir/vol'

        exp_ret_val = {
            'reference': {'source-name': '//host/share/subdir/vol'},
            'size': exp_size,
            'safe_to_manage': exp_manageable,
            'reason_not_safe': mock.ANY,
            'cinder_id': managed_volume.id if managed_volume else None,
            'extra_info': None,
        }

        mock_get_size.side_effect = get_size_side_effect
        mock_check_manageable.return_value = (manageable_check_ret_val,
                                              mock.sentinel.reason)
        mock_get_mount_point.return_value = mountpoint

        ret_val = self._driver._get_manageable_volume(
            share, volume_path, managed_volume)
        self.assertEqual(exp_ret_val, ret_val)

        mock_check_manageable.assert_called_once_with(
            volume_path, already_managed=managed_volume is not None)
        mock_get_mount_point.assert_called_once_with(share)
        if managed_volume:
            mock_get_size.assert_not_called()
        else:
            mock_get_size.assert_called_once_with(volume_path)

    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       '_get_mount_point_for_share', create=True)
    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       '_get_manageable_volume')
    @mock.patch.object(os, 'walk')
    @mock.patch.object(os.path, 'join', lambda *args: '/'.join(args))
    def test_get_share_manageable_volumes(
            self, mock_walk, mock_get_manageable_volume,
            mock_get_mount_point):
        mount_path = '/fake-mountpoint'
        mock_walk.return_value = [
            [mount_path, ['subdir'], ['volume-1.vhdx']],
            ['/fake-mountpoint/subdir', [], ['volume-0', 'volume-3.vhdx']]]

        mock_get_manageable_volume.side_effect = [
            Exception,
            mock.sentinel.managed_volume]

        self._driver._MANAGEABLE_IMAGE_RE = re.compile(r'.*\.(?:vhdx)$')

        managed_volumes = {'volume-1': mock.sentinel.vol1}

        exp_manageable = [mock.sentinel.managed_volume]
        manageable_volumes = self._driver._get_share_manageable_volumes(
            mock.sentinel.share,
            managed_volumes)

        self.assertEqual(exp_manageable, manageable_volumes)

        mock_get_manageable_volume.assert_has_calls(
            [mock.call(mock.sentinel.share,
                       '/fake-mountpoint/volume-1.vhdx',
                       mock.sentinel.vol1),
             mock.call(mock.sentinel.share,
                       '/fake-mountpoint/subdir/volume-3.vhdx',
                       None)])

    @mock.patch.object(remotefs.RemoteFSManageableVolumesMixin,
                       '_get_share_manageable_volumes')
    @mock.patch.object(volume_utils, 'paginate_entries_list')
    def test_get_manageable_volumes(self, mock_paginate, mock_get_share_vols):
        fake_vol = fake_volume.fake_volume_obj(mock.sentinel.context)
        self._driver._mounted_shares = [mock.sentinel.share0,
                                        mock.sentinel.share1]

        mock_get_share_vols.side_effect = [
            Exception, [mock.sentinel.manageable_vol]]

        pagination_args = [
            mock.sentinel.marker, mock.sentinel.limit,
            mock.sentinel.offset, mock.sentinel.sort_keys,
            mock.sentinel.sort_dirs]
        ret_val = self._driver.get_manageable_volumes(
            [fake_vol], *pagination_args)

        self.assertEqual(mock_paginate.return_value, ret_val)
        mock_paginate.assert_called_once_with(
            [mock.sentinel.manageable_vol], *pagination_args)

        exp_managed_vols_dict = {fake_vol.name: fake_vol}
        mock_get_share_vols.assert_has_calls(
            [mock.call(share, exp_managed_vols_dict)
             for share in self._driver._mounted_shares])
